#ifndef _core_h_
#define _core_h_

#include "util.h"

typedef enum
{
	GHANDLER_NONE=0,
	GHANDLER_ROOT,
	GHANDLER_INTRO,
	GHANDLER_MENU,
	GHANDLER_CREDITS,
	GHANDLER_OPTIONS,
	GHANDLER_HELP,
	GHANDLER_GAME,
	GHANDLER_MAZE,
	GHANDLER_INVENTORY,
	GHANDLER_FILE_MANAGER,
	GHANDLER_EPILOG,
	GHANDLER_PAUSED
} game_handlers;

typedef enum
{
	TIMER_WALL,
	TIMER_DOOR,
	TIMER_PIT,
	TIMER_TELEPORTATION,
	TIMER_CHARACTER,
	TIMER_LIGHT
} timer_type;

struct timer_event
{
	game_handlers	game_handler;		/* parent game handler */
	Uint32			ticks;				/* last ticks count */
	Uint32			ticks_period;		/* minimal ticks in one periode */
	bool			expires;			/* use the 'remaining' attribute? */
	Uint32			remaining;			/* number of remaining periods (count-down) */
	Uint32			elapsed_periods;	/* elapsed periods number */
	timer_type		type;				/* timered event type */
	ptr				data;				/* generic data pointer */
};

typedef struct timer_event* p_timer_event;


void			commandLine_error(const str msg, const str arg);
void			missing_argument(const str arg);
void			commandLine(const int argc, const str argv[]);
void			core_init(void);
void			core_reset(void);
void			fast_quit(void);

void			push_game_handler(const game_handlers game_handler);
void			pop_game_handler(void);
game_handlers	get_current_game_handler(void);
void			reset_game_handler_stack(void);

bool			timers_scheduler(void);
void			timers_register_new(const Uint32 period, const bool expires, const Uint32 remaining,
							const timer_type type, const ptr data, const bool exclusive);
void			timers_compute_data(p_timer_event timer);
void			timers_remove_data(p_timer_event timer);
void			timers_flush(void);
void			timers_flush_filtered(const game_handlers game_handler);
void			timers_lock(void);
void			timers_unlock(void);

#endif
