#include "game.h"
#include "core.h"
#include "gfx.h"
#include "IO.h"


extern bool					redraw;
extern bool					draw_alternate;
extern game_handlers		game_handler;
extern SDL_Surface*			screen;
extern struct video_options	video_conf;
/*extern struct mouse_status mouse;
extern struct video_options video_conf;*/

p_map				maps_table[MAPS];
struct team_set		team;
character_motion	keypad_set;
character_motion	keypad_old;


/****************************************************************************/
/*									TABLES									*/
/****************************************************************************/


struct movement movements_table[]=
/* translation table = (going_relative, facing_absolute)=going_absolute */
{
	{ CMOTION_GOING_FRONT,	CFACING_NORTH,	CMOVE_GOING_NORTH },
	{ CMOTION_GOING_FRONT,	CFACING_EAST,	CMOVE_GOING_EAST  },
	{ CMOTION_GOING_FRONT,	CFACING_SOUTH,	CMOVE_GOING_SOUTH },
	{ CMOTION_GOING_FRONT,	CFACING_WEST,	CMOVE_GOING_WEST  },
	{ CMOTION_TURNING_LEFT,	CFACING_WEST,	CMOVE_GOING_WEST  },

	{ CMOTION_GOING_RIGHT,	CFACING_NORTH,	CMOVE_GOING_EAST  },
	{ CMOTION_GOING_RIGHT,	CFACING_EAST,	CMOVE_GOING_SOUTH },
	{ CMOTION_GOING_RIGHT,	CFACING_SOUTH,	CMOVE_GOING_WEST  },
	{ CMOTION_GOING_RIGHT,	CFACING_WEST,	CMOVE_GOING_NORTH },

	{ CMOTION_GOING_BACK,	CFACING_NORTH,	CMOVE_GOING_SOUTH },
	{ CMOTION_GOING_BACK,	CFACING_EAST,	CMOVE_GOING_WEST  },
	{ CMOTION_GOING_BACK,	CFACING_SOUTH,	CMOVE_GOING_NORTH },
	{ CMOTION_GOING_BACK,	CFACING_WEST,	CMOVE_GOING_EAST  },

	{ CMOTION_GOING_LEFT,	CFACING_NORTH,	CMOVE_GOING_WEST  },
	{ CMOTION_GOING_LEFT,	CFACING_EAST,	CMOVE_GOING_NORTH },
	{ CMOTION_GOING_LEFT,	CFACING_SOUTH,	CMOVE_GOING_EAST  },
	{ CMOTION_GOING_LEFT,	CFACING_WEST,	CMOVE_GOING_SOUTH },

	{ 0, 0, 0 }
};
	
str directions_table[]=
{
	"NORTH",
	"EAST",
	"SOUTH",
	"WEST",
	NULL
};
	
#ifdef DEBUG
map_element default_map_0[20*20]=
{
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 2, 0, 1,
	1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 2, 1,
	1, 0, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 1, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
	1, 3, 1, 3, 1, 3, 1, 3, 1, 1, 6, 0, 0, 0, 1, 0, 0, 0, 1, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
};

map_element default_map_1[5*5]=
{
	1, 1, 1, 1, 1,
	1, 0, 0, 0, 1,
	1, 0, 0, 0, 1,
	1, 0, 0, 0, 4,
	1, 1, 1, 1, 1,
};

struct map default_maps_table[2]=
{	
	{
		"map #1 (default)",
		0,
		20, 20, 1, 10, CFACING_NORTH,
		default_map_0,
		NULL,
		NULL
	},
	{
		"map #2",
		0,
		5, 5, 1, 1, CFACING_NORTH,
		default_map_1,
		NULL,
		NULL
	}
};

struct map_button default_buttons_table[1]=
{
	{	0,	ITEM_BUTTON_0,	true,	true	}
};

struct map_wall default_walls_table[2]=
{
	{	0,	ITEM_STONE_WALL,		true,	false,	0,	NULL, NULL, },
	{	1,	ITEM_INVISIBLE_WALL,	false,	true,	0,	NULL, NULL, }
};

struct map_door default_doors_table[5]=
{
	{	0,	ITEM_IRON_DOOR,		CFACING_NORTH,	DOOR_OPENED,		DMOTION_NONE,	true,	NULL, NULL, NULL, NULL, NULL, },
	{	1,	ITEM_WOODEN_DOOR,	CFACING_NORTH,	DOOR_NEARLY_OPENED,	DMOTION_NONE,	true,	NULL, NULL, NULL, NULL, NULL, },
	{	2,	ITEM_IRON_DOOR,		CFACING_NORTH,	DOOR_HALF_CLOSED,	DMOTION_NONE,	true,	NULL, NULL, NULL, NULL, NULL, },
	{	3,	ITEM_IRON_DOOR,		CFACING_NORTH,	DOOR_NEARLY_CLOSED,	DMOTION_NONE,	true,	NULL, NULL, NULL, NULL, NULL, },
	{	4,	ITEM_IRON_DOOR,		CFACING_EAST,	DOOR_CLOSED,		DMOTION_NONE,	true,	NULL, NULL, NULL, NULL, NULL, }
};

struct map_stairs default_stairs_table[2]=
{
	{	0,	ITEM_STAIRS_UP,		CFACING_WEST,	 3,	 3,	1	},
	{	0,	ITEM_STAIRS_DOWN,	CFACING_WEST,	18,	18,	0	}
};

struct map_pit default_pits_table[1]=
{
	{	0,	ITEM_PIT_0,	true,	true,	CFACING_NEUTRAL,	2,	2,	1	}
};
#endif


/****************************************************************************/
/*									MAP I/O									*/
/****************************************************************************/


#ifdef DEBUG
void map_load_default_maps(void)
{
	p_map	themap;
	int		i, field_size, stairs, doors, pits, walls;

	walls=0;
	pits=0;
	stairs=0;
	doors=0;

	themap=&default_maps_table[0];
	field_size=themap->xsize*themap->ysize;
	themap->elements=(ptr*)malloc(field_size*sizeof(map_element));
	if (themap->elements==NULL)
		fatal("game.map", "couldn't init default map buffers");
	for (i=0; i<field_size; i++)
	{
		switch (themap->tiles[i])
		{
			case MAP_MAGICAL_WALL:
				themap->elements[i]=(ptr)&default_walls_table[walls++];
				break;
			case MAP_DOOR:
				if (doors==0)
				{
					default_doors_table[doors].buttons=(p_list)malloc(sizeof(struct list));
					if (default_doors_table[doors].buttons==NULL)
						fatal("game.load_default_maps", "cannot allocate buffer for an item list structure");
					list_init(default_doors_table[doors].buttons);
					list_push(default_doors_table[doors].buttons, (ptr)&default_buttons_table[0]);
				}
				themap->elements[i]=(ptr)&default_doors_table[doors++];
				break;
			case MAP_STAIRS:
				themap->elements[i]=(ptr)&default_stairs_table[stairs++];
				break;
			case MAP_PIT:
				themap->elements[i]=(ptr)&default_pits_table[pits++];
				break;
			default:
				themap->elements[i]=NULL;
		}
	}
	maps_table[0]=themap;

	themap=&default_maps_table[1];
	field_size=themap->xsize*themap->ysize;
	themap->elements=(ptr*)malloc(field_size*sizeof(map_element));
	if (themap->elements==NULL)
		fatal("game.map", "couldn't init default map buffers");
	for (i=0; i<field_size; i++)
	{
		switch (themap->tiles[i])
		{
			case MAP_MAGICAL_WALL:
				themap->elements[i]=(ptr)&default_walls_table[walls++];
				break;
			case MAP_DOOR:
				themap->elements[i]=(ptr)&default_doors_table[doors++];
				break;
			case MAP_STAIRS:
				themap->elements[i]=(ptr)&default_stairs_table[stairs++];
				break;
			case MAP_PIT:
				themap->elements[i]=(ptr)&default_pits_table[pits++];
				break;
			default:
				themap->elements[i]=NULL;
		}
	}
	maps_table[1]=themap;
}
#endif

#define fload(X,Y) \
	{\
		ret=fread(X, Y, 1, in); \
		if (ret!=1) \
			error("game.map", "write error");\
	}


void map_load(const str datafile, const int map_index)
/* TODO: add error checking */
{
	fileHandle	in;
	char		tmp[STR_MAX];
	size_t		ret;
	int			i, field_size;
	ulong		cnt;
	p_map		themap;

	sprintf(tmp, "loading map #%d..", map_index);
	info("game.map", tmp);

	map_delete(map_index);

	in=fopen(datafile, "rb");
	if ((in==NULL) OR (datafile==NULL))
	{
		sprintf(tmp, "cannot load map from file %s", datafile);
		error("game.map", tmp);
	}

	themap=(p_map)malloc(sizeof(struct map));
	if (themap==NULL)
		fatal("game.map", "cannot allocate main buffer");

	themap->name=NULL;
	themap->tiles=NULL;
	themap->elements=NULL;
	themap->characters=NULL;
	
	/* name */
	fload(&cnt, sizeof(cnt));
	themap->name=(str)malloc(cnt+1);
	if (themap->name==NULL)
		fatal("game.map", "cannot allocate buffer for attribute 'name'");
	if (cnt)
		fload(themap->name, cnt);
	themap->name[cnt]='\0';

	/* other attributes */
	fload(&themap->level, sizeof(themap->level));
	fload(&themap->xsize, sizeof(themap->xsize));
	fload(&themap->ysize, sizeof(themap->ysize));
	fload(&themap->initial_x, sizeof(themap->initial_x));
	fload(&themap->initial_y, sizeof(themap->initial_y));
	fload(&themap->initial_facing, sizeof(themap->initial_facing));
	field_size=themap->xsize*themap->ysize;

	/* load the map_element field */
	themap->tiles=(map_element*)malloc(field_size*sizeof(map_element));
	if (themap->tiles==NULL)
		fatal("game.map", "cannot allocate buffer for attribute 'tile'");
	fload(themap->tiles, field_size*sizeof(map_element));

	/* load each map_element structure implementation */
	themap->elements=(ptr*)malloc(field_size*sizeof(ptr));
	if (themap->elements==NULL)
		fatal("game.map", "cannot allocate buffer for attribute 'elements'");
	for (i=0; i<field_size; i++)
	{
		themap->elements[i]=NULL;
		fload(&cnt, sizeof(cnt));
		if (cnt==1)
			map_load_element(in, themap->tiles[i], &themap->elements[i]);
	}

	/* load the characters list */
	map_load_character_list(in, &themap->characters);

	fclose(in);

	maps_table[map_index]=themap;

	info("game.map", "map loaded");
}

void map_load_element(const fileHandle in, const map_element type, ptr* data)
{
	p_map_ground			ground;
	p_map_wall				wall;
	p_map_door				door;
	p_map_stairs			stairs;
	p_map_pit				pit;
	p_map_trap				trap;
	p_map_teleportation		teleportation;
	p_map_button			button;
	size_t					ret;
	char					tmp[STR_MAX];
	map_element				effective_type;

	if (data==NULL)
		fatal("game.map", "attempting to load a NULL element");

	if (*data)
	{
		error("game.map", "attempting to overload an element");
		return;
	}

	fload(&effective_type, sizeof(effective_type));
	if (effective_type!=type)
	{
		error("game.map", "unexpected map element type, map file may be damaged");
		return;
	}
	switch (type)
	{
		case MAP_GROUND:
			ground=(p_map_ground)malloc(sizeof(struct map_ground));
			if (ground==NULL)
				fatal("game.map", "cannot allocate buffer for map element's 'ground'");
			ground->itemsNE=NULL;
			ground->itemsNW=NULL;
			ground->itemsSE=NULL;
			ground->itemsSW=NULL;
			fload(&ground->UID, sizeof(ground->UID));
			fload(&ground->decoration, sizeof(ground->decoration));
			map_load_item_list(in, &ground->itemsNE);
			map_load_item_list(in, &ground->itemsNW);
			map_load_item_list(in, &ground->itemsSE);
			map_load_item_list(in, &ground->itemsSW);
			*data=ground;
			break;
		case MAP_MAGICAL_WALL:
		case MAP_WALL:
			wall=(p_map_wall)malloc(sizeof(struct map_wall));
			if (wall==NULL)
				fatal("game.map", "cannot allocate buffer for map element's 'wall'");
			wall->buttons=NULL;
			wall->items=NULL;
			fload(&wall->UID, sizeof(wall->UID));
			fload(&wall->type, sizeof(wall->type));
			fload(&wall->enabled, sizeof(wall->enabled));
			fload(&wall->visible, sizeof(wall->visible));
			fload(&wall->decoration, sizeof(wall->decoration));
			map_load_element_list(in, MAP_BUTTON, &wall->buttons);
			map_load_item_list(in, &wall->items);
			*data=wall;
			break;
		case MAP_DOOR:
			door=(p_map_door)malloc(sizeof(struct map_door));
			if (door==NULL)
				fatal("game.map", "cannot allocate buffer for map element's 'door'");
			door->buttons=NULL;
			door->itemsNE=NULL;
			door->itemsNW=NULL;
			door->itemsSE=NULL;
			door->itemsSW=NULL;
			fload(&door->UID, sizeof(door->UID));
			fload(&door->type, sizeof(door->type));
			fload(&door->orientation, sizeof(door->orientation));
			fload(&door->stage, sizeof(door->stage));
			fload(&door->motion, sizeof(door->motion));
			fload(&door->enabled, sizeof(door->enabled));
			map_load_element_list(in, MAP_BUTTON, &door->buttons);
			map_load_item_list(in, &door->itemsNE);
			map_load_item_list(in, &door->itemsNW);
			map_load_item_list(in, &door->itemsSE);
			map_load_item_list(in, &door->itemsSW);
			*data=door;
			break;
		case MAP_STAIRS:
			stairs=(p_map_stairs)malloc(sizeof(struct map_stairs));
			if (stairs==NULL)
				fatal("game.map", "cannot allocate buffer for map element's 'stairs'");
			fload(&stairs->UID, sizeof(stairs->UID));
			fload(&stairs->type, sizeof(stairs->type));
			fload(&stairs->orientation, sizeof(stairs->orientation));
			fload(&stairs->x, sizeof(stairs->x));
			fload(&stairs->y, sizeof(stairs->y));
			fload(&stairs->level, sizeof(stairs->level));
			*data=stairs;
			break;
		case MAP_TRAP:
			trap=(p_map_trap)malloc(sizeof(struct map_trap));
			if (trap==NULL)
				fatal("game.map", "cannot allocate buffer for map element's 'trap'");
			fload(&trap->UID, sizeof(trap->UID));
			fload(&trap->type, sizeof(trap->type));
			fload(&trap->enabled, sizeof(trap->enabled));
			fload(&trap->visible, sizeof(trap->visible));
			*data=trap;
			break;
		case MAP_PIT:
			pit=(p_map_pit)malloc(sizeof(struct map_pit));
			if (pit==NULL)
				fatal("game.map", "cannot allocate buffer for map element's 'pit'");
			fload(&pit->UID, sizeof(pit->UID));
			fload(&pit->type, sizeof(pit->type));
			fload(&pit->enabled, sizeof(pit->enabled));
			fload(&pit->visible, sizeof(pit->visible));
			fload(&pit->orientation, sizeof(pit->orientation));
			fload(&pit->x, sizeof(pit->x));
			fload(&pit->y, sizeof(pit->y));
			fload(&pit->level, sizeof(pit->level));
			*data=pit;
			break;
		case MAP_TELEPORTATION:
			teleportation=(p_map_teleportation)malloc(sizeof(struct map_teleportation));
			if (teleportation==NULL)
				fatal("game.map", "cannot allocate buffer for map element's 'teleportation'");
			fload(&teleportation->UID, sizeof(teleportation->UID));
			fload(&teleportation->type, sizeof(teleportation->type));
			fload(&teleportation->enabled, sizeof(teleportation->enabled));
			fload(&teleportation->visible, sizeof(teleportation->visible));
			fload(&teleportation->orientation, sizeof(teleportation->orientation));
			fload(&teleportation->x, sizeof(teleportation->x));
			fload(&teleportation->y, sizeof(teleportation->y));
			fload(&teleportation->level, sizeof(teleportation->level));
			*data=teleportation;
			break;
		case MAP_BUTTON:
			button=(p_map_button)malloc(sizeof(struct map_button));
			if (button==NULL)
				fatal("game.map", "cannot allocate buffer for map element's 'button'");
			fload(&button->UID, sizeof(button->UID));
			fload(&button->type, sizeof(button->type));
			fload(&button->enabled, sizeof(button->enabled));
			fload(&button->visible, sizeof(button->visible));
			*data=button;
			break;
		default:
			sprintf(tmp, "unexpected element %d to load", type);
			error("game.map", tmp);
			*data=NULL;
	}
}

void map_load_element_list(const fileHandle in, const map_element type, p_list* element_list)
{
	map_element*	content;
	size_t			ret;
	p_list			thelist;
	ulong			i, cnt;

	if (element_list==NULL)
		fatal("game.map", "attempting to load a NULL element list");

	if (*element_list)
	{
		error("game.map", "attempting to overload an element list");
		return;
	}

	thelist=(p_list)malloc(sizeof(struct list));
	if (thelist==NULL)
		fatal("game.map", "cannot allocate buffer for an element_list list structure");
	list_init(thelist);
	
	fload(&cnt, sizeof(cnt));
	for (i=0; i<cnt; i++)
	{
		content=NULL;
		map_load_element(in, type, (ptr)&content);
		list_push(thelist, content);
	}
	if (cnt!=thelist->size)
		error("game.map", "list size doesn't match the number of elements loaded");

	*element_list=thelist;
}

void map_load_character(const fileHandle in, p_map_character* data)
{
	p_map_character	thecharacter;
	ulong			cnt;
	size_t			ret;

	if (data==NULL)
		fatal("game.map", "attempting to load a NULL character");

	if (*data)
	{
		error("game.map", "attempting to overload a character structure");
		return;
	}

	thecharacter=(p_map_character)malloc(sizeof(struct map_character));
	if (thecharacter==NULL)
		fatal("game.map", "cannot allocate buffer for a character structure");
	
	thecharacter->items=NULL;
	
	fload(&thecharacter->UID, sizeof(thecharacter->UID));
	fload(&thecharacter->type, sizeof(thecharacter->type));

	/* name */
	fload(&cnt, sizeof(cnt));
	if (cnt)
	{
		thecharacter->name=(str)malloc(cnt+1);
		if (thecharacter->name==NULL)
			fatal("game.map", "cannot allocate buffer for character attribute 'name'");
		fload(thecharacter->name, cnt);
		thecharacter->name[cnt]='\0';
	}
	else
		thecharacter->name="";

	/* TODO: complete */

	*data=thecharacter;
}

void map_load_character_list(const fileHandle in, p_list* character_list)
{
	p_map_character	content;
	size_t			ret;
	p_list			thelist;
	ulong			i, cnt;

	if (character_list==NULL)
		fatal("game.map", "attempting to load a NULL character list");

	if (*character_list)
	{
		error("game.map", "attempting to overload a character list");
		return;
	}

	thelist=(p_list)malloc(sizeof(struct list));
	if (thelist==NULL)
		fatal("game.map", "cannot allocate buffer for a character list structure");
	list_init(thelist);
	
	fload(&cnt, sizeof(cnt));
	for (i=0; i<cnt; i++)
	{
		content=NULL;
		map_load_character(in, &content);
		list_push(thelist, content);
	}
	if (cnt!=thelist->size)
		error("game.map", "list size doesn't match the number of characters loaded");

	*character_list=thelist;
}

void map_load_item(const fileHandle in, p_map_item* data)
{
	p_map_item	theitem;
	ulong		cnt;
	size_t		ret;

	if (data==NULL)
		fatal("game.map", "attempting to load a NULL item");

	if (*data)
	{
		error("game.map", "attempting to overload an item structure");
		return;
	}

	theitem=(p_map_item)malloc(sizeof(struct map_item));
	if (theitem==NULL)
		fatal("game.map", "cannot allocate buffer for an item structure");
	theitem->items=NULL;

	fload(&theitem->UID, sizeof(theitem->UID));
	fload(&theitem->type, sizeof(theitem->type));

	/* name */
	fload(&cnt, sizeof(cnt));
	if (cnt)
	{
		theitem->name=(str)malloc(cnt+1);
		if (theitem->name==NULL)
			fatal("game.map", "cannot allocate buffer for item attribute 'name'");
		fload(theitem->name, cnt);
		theitem->name[cnt]='\0';
	}
	else
		theitem->name="";
	fload(&theitem->weight, sizeof(theitem->weight));

	/* TODO: complete */

	*data=theitem;
}

void map_load_item_list(const fileHandle in, p_list* item_list)
{
	p_map_item	content;
	size_t		ret;
	p_list		thelist;
	ulong		i, cnt;

	if (item_list==NULL)
		fatal("game.map", "attempting to load a NULL item list");

	if (*item_list)
	{
		error("game.map", "attempting to overload an item list");
		return;
	}

	thelist=(p_list)malloc(sizeof(struct list));
	if (thelist==NULL)
		fatal("game.map", "cannot allocate buffer for an item list structure");

	list_init(thelist);

	fload(&cnt, sizeof(cnt));
	for (i=0; i<cnt; i++)
	{
		content=NULL;
		map_load_item(in, &content);
		list_push(thelist, content);
	}
	if (cnt!=thelist->size)
		error("game.map", "list size doesn't match the number of items loaded");

	*item_list=thelist;
}

#define fsave(X,Y) \
	{\
		ret=fwrite(X, Y, 1, out); \
		if (ret!=1) \
			error("game.map", "write error");\
	}

void map_save(const str datafile, const int map_index)
/* TODO: add error checking */
{
	fileHandle	out;
	char		tmp[STR_MAX];
	size_t		ret;
	int			i, field_size;
	ulong		cnt;
	p_map		themap;

	sprintf(tmp, "saving map #%d..", map_index);
	info("game.map", tmp);

	themap=maps_table[map_index];

	if (themap==NULL)
		fatal("game.map", "attempting to save a NULL map");
		
	out=fopen(datafile, "wb");
	if ((out==NULL) OR (datafile==NULL))
	{
		sprintf(tmp, "cannot save map to file %s", datafile);
		error("game.map", tmp);
	}

	/* store attributes */
	if (themap->name)
	{
		cnt=strlen(themap->name);
		fsave(&cnt, sizeof(cnt));
		fsave(themap->name, cnt);
	}
	else
	{
		cnt=0;
		fsave(&cnt, sizeof(cnt));
	}
	fsave(&themap->level, sizeof(themap->level));
	fsave(&themap->xsize, sizeof(themap->xsize));
	fsave(&themap->ysize, sizeof(themap->ysize));
	fsave(&themap->initial_x, sizeof(themap->initial_x));
	fsave(&themap->initial_y, sizeof(themap->initial_y));
	fsave(&themap->initial_facing, sizeof(themap->initial_facing));
	field_size=themap->xsize*themap->ysize;

	/* store the map_element field */
	if (themap->tiles==NULL)
	{
		error("game.map", "can't save: map has no tiles");
		return;
	}
	fsave(&themap->tiles[0], field_size*sizeof(themap->tiles[0]));

	/* store each map_element structure implementation */
	if (themap->elements==NULL)
	{
		error("game.map", "can't save: map has no elements");
		return;
	}
	for (i=0; i<field_size; i++)
	{
		if (themap->elements[i])
		{
			cnt=1;
			fsave(&cnt, sizeof(cnt));
			map_save_element(out, themap->tiles[i], themap->elements[i]);
		}
		else
		{
			cnt=0;
			fsave(&cnt, sizeof(cnt));
		}
	}

	/* store the characters list */
	map_save_character_list(out, themap->characters);

	fclose(out);

	info("game.map", "map saved");
}

void map_save_element(const fileHandle out, const map_element type, ptr data)
{
	p_map_ground			ground;
	p_map_wall				wall;
	p_map_door				door;
	p_map_stairs			stairs;
	p_map_pit				pit;
	p_map_trap				trap;
	p_map_teleportation		teleportation;
	p_map_button			button;
	size_t					ret;
	char					tmp[STR_MAX];

	if (data==NULL)
	{
		error("game.map", "attempting to save a NULL element structure");
		return;
	}

	fsave(&type, sizeof(type));	/* save the type (used to check integrity when loading) */
	switch (type)
	{
		case MAP_GROUND:
			ground=(p_map_ground)data;
			fsave(&ground->UID, sizeof(ground->UID));
			fsave(&ground->decoration, sizeof(ground->decoration));
			map_save_item_list(out, ground->itemsNE);
			map_save_item_list(out, ground->itemsNW);
			map_save_item_list(out, ground->itemsSE);
			map_save_item_list(out, ground->itemsSW);
			break;
		case MAP_MAGICAL_WALL:
		case MAP_WALL:
			wall=(p_map_wall)data;
			fsave(&wall->UID, sizeof(wall->UID));
			fsave(&wall->type, sizeof(wall->type));
			fsave(&wall->enabled, sizeof(wall->enabled));
			fsave(&wall->visible, sizeof(wall->visible));
			fsave(&wall->decoration, sizeof(wall->decoration));
			map_save_element_list(out, MAP_BUTTON, wall->buttons);
			map_save_item_list(out, wall->items);
			break;
		case MAP_DOOR:
			door=(p_map_door)data;
			fsave(&door->UID, sizeof(door->UID));
			fsave(&door->type, sizeof(door->type));
			fsave(&door->orientation, sizeof(door->orientation));
			fsave(&door->stage, sizeof(door->stage));
			fsave(&door->motion, sizeof(door->motion));
			fsave(&door->enabled, sizeof(door->enabled));
			map_save_element_list(out, MAP_BUTTON, door->buttons);
			map_save_item_list(out, door->itemsNE);
			map_save_item_list(out, door->itemsNW);
			map_save_item_list(out, door->itemsSE);
			map_save_item_list(out, door->itemsSW);
			break;
		case MAP_STAIRS:
			stairs=(p_map_stairs)data;
			fsave(&stairs->UID, sizeof(stairs->UID));
			fsave(&stairs->type, sizeof(stairs->type));
			fsave(&stairs->orientation, sizeof(stairs->orientation));
			fsave(&stairs->x, sizeof(stairs->x));
			fsave(&stairs->y, sizeof(stairs->y));
			fsave(&stairs->level, sizeof(stairs->level));
			break;
		case MAP_TRAP:
			trap=(p_map_trap)data;
			fsave(&trap->UID, sizeof(trap->UID));
			fsave(&trap->type, sizeof(trap->type));
			fsave(&trap->enabled, sizeof(trap->enabled));
			fsave(&trap->visible, sizeof(trap->visible));
			break;
		case MAP_PIT:
			pit=(p_map_pit)data;
			fsave(&pit->UID, sizeof(pit->UID));
			fsave(&pit->type, sizeof(pit->type));
			fsave(&pit->enabled, sizeof(pit->enabled));
			fsave(&pit->visible, sizeof(pit->visible));
			fsave(&pit->orientation, sizeof(pit->orientation));
			fsave(&pit->x, sizeof(pit->x));
			fsave(&pit->y, sizeof(pit->y));
			fsave(&pit->level, sizeof(pit->level));
			break;
		case MAP_TELEPORTATION:
			teleportation=(p_map_teleportation)data;
			fsave(&teleportation->UID, sizeof(teleportation->UID));
			fsave(&teleportation->type, sizeof(teleportation->type));
			fsave(&teleportation->enabled, sizeof(teleportation->enabled));
			fsave(&teleportation->visible, sizeof(teleportation->visible));
			fsave(&teleportation->orientation, sizeof(teleportation->orientation));
			fsave(&teleportation->x, sizeof(teleportation->x));
			fsave(&teleportation->y, sizeof(teleportation->y));
			fsave(&teleportation->level, sizeof(teleportation->level));
			break;
		case MAP_BUTTON:
			button=(p_map_button)data;
			fsave(&button->UID, sizeof(button->UID));
			fsave(&button->type, sizeof(button->type));
			fsave(&button->enabled, sizeof(button->enabled));
			fsave(&button->visible, sizeof(button->visible));
			break;
		default:
			sprintf(tmp, "unexpected element %d to save", type);
			warning("game.map", tmp);
			NOP;
	}
}

void map_save_element_list(const fileHandle out, const map_element type, p_list element_list)
{
	p_cell	c;
	size_t	ret;
	ulong	cnt;
	
	if (element_list==NULL)
		cnt=0;
	else
		cnt=element_list->size;
	fsave(&cnt, sizeof(cnt));

	if (element_list==NULL)
		return;

	cnt=0;
	c=element_list->head;
	while (c)
	{
		if (c->content)
		{
			map_save_element(out, type, (ptr)c->content);
			cnt++;
		}
		c=c->next;
	}
	if (cnt!=element_list->size)
		error("game.map", "list size doesn't match the number of elements saved");
}

void map_save_character(const fileHandle out, p_map_character data)
{
	ulong	cnt;
	size_t	ret;

	if (data==NULL)
	{
		error("game.map", "attempting to save a NULL character structure");
		return;
	}

	fsave(&data->UID, sizeof(data->UID));
	fsave(&data->type, sizeof(data->type));
	if (data->name)
	{
		cnt=strlen(data->name);
		fsave(&cnt, sizeof(cnt));
		fsave(&data->name, cnt);
	}
	else
	{
		cnt=0;
		fsave(&cnt, sizeof(cnt));
	}
	/* TODO: complete */
}

void map_save_character_list(const fileHandle out, p_list character_list)
{
	p_cell	c;
	size_t	ret;
	ulong	cnt;
	
	if (character_list==NULL)
		cnt=0;
	else
		cnt=character_list->size;
	fsave(&cnt, sizeof(cnt));

	if (character_list==NULL)
		return;

	cnt=0;
	c=character_list->head;
	while (c)
	{
		if (c->content)
		{
			map_save_character(out, (p_map_character)c->content);
			cnt++;
		}
		c=c->next;
	}
	if (cnt!=character_list->size)
		error("game.map", "list size doesn't match the number of characters saved");
}

void map_save_item(const fileHandle out, p_map_item data)
{
	ulong	cnt;
	size_t	ret;

	if (data==NULL)
	{
		error("game.map", "attempting to save a NULL item structure");
		return;
	}

	fsave(&data->UID, sizeof(data->UID));
	fsave(&data->type, sizeof(data->type));
	if (data->name)
	{
		cnt=strlen(data->name);
		fsave(&cnt, sizeof(cnt));
		fsave(&data->name, cnt);
	}
	else
	{
		cnt=0;
		fsave(&cnt, sizeof(cnt));
	}
	fsave(&data->weight, sizeof(data->weight));
	/* TODO: complete */
}

void map_save_item_list(const fileHandle out, p_list item_list)
{
	p_cell	c;
	size_t	ret;
	ulong	cnt;

	if (item_list==NULL)
		cnt=0;
	else
		cnt=item_list->size;
	fsave(&cnt, sizeof(cnt));
	
	if (item_list==NULL)
		return;

	cnt=0;
	c=item_list->head;
	while (c)
	{
		if (c->content)
		{
			map_save_item(out, (p_map_item)c->content);
			cnt++;
		}
		c=c->next;
	}
	if (cnt!=item_list->size)
		error("game.map", "list size doesn't match the number of items saved");
}	
		
void map_get_tile(int x, int y, map_element* tile, ptr* element)
{
	ulong index;

	if (	(x<0)
		OR	(x>=team.map->xsize)
		OR	(y<0)
		OR	(y>=team.map->ysize))
	{
/*		char tmp[STR_MAX];

*/		*tile=MAP_WALL;
		*element=NULL;
/*		sprintf(tmp, "requesting tile (%d,%d): outside of map limits", x, y);
		warning("game.map", tmp);
*/		return;
	}
	index=y*team.map->ysize+x;
	*tile=team.map->tiles[index];
	*element=team.map->elements[index];
}

void map_get_tile_relative(int front_switch, int side_switch, map_element* tile, ptr* element)
{
	int	x, y;

	x=team.position.x;
	y=team.position.y;
	switch (team.position.facing)
	{
		case CFACING_NORTH:
			x+=front_switch;
			y+=side_switch;
			break;
		case CFACING_EAST:
			x-=side_switch;
			y+=front_switch;
			break;
		case CFACING_SOUTH:
			x-=front_switch;
			y-=side_switch;
			break;
		case CFACING_WEST:
			x+=side_switch;
			y-=front_switch;
			break;
		default:
			NOP
	}
	map_get_tile(x, y, tile, element);
}

void map_delete(const int map_index)
{
	p_map	themap;
	int		field_size, i;

	themap=maps_table[map_index];
	if (themap==NULL)
		return;

	/* delete the map name */
	xfree((ptr)&themap->name);

	/* delete the elements arrays */
	field_size=themap->xsize*themap->ysize;
	for (i=0; i<field_size; i++)
		if (themap->elements[i])
			map_delete_element(themap->tiles[i], &themap->elements[i]);
	xfree((ptr)&themap->elements);
	xfree((ptr)&themap->tiles);

	/* delete the characters list */
	map_delete_character_list(&themap->characters);

	/* delete the map itself (and its entry in maps_table) */
	xfree((ptr)&themap);
	maps_table[map_index]=NULL;
}

void map_delete_element(const map_element type, ptr* data)
/* free an element structure and resets its pointer */
{
	p_map_ground			ground;
	p_map_wall				wall;
	p_map_door				door;
	p_map_stairs			stairs;
	p_map_pit				pit;
	p_map_trap				trap;
	p_map_teleportation		teleportation;
	p_map_button			button;
	char					tmp[STR_MAX];

	if (data==NULL)
	{
		error("game.map", "attempting to delete a NULL element pointer");
		return;
	}

	if (*data==NULL)
	{
		error("game.map", "attempting to delete a NULL element structure");
		return;
	}
		
	switch (type)
	{
		case MAP_GROUND:
			ground=(p_map_ground)*data;
			map_delete_item_list(&ground->itemsNE);
			map_delete_item_list(&ground->itemsNW);
			map_delete_item_list(&ground->itemsSE);
			map_delete_item_list(&ground->itemsSW);
			break;
		case MAP_MAGICAL_WALL:
		case MAP_WALL:
			wall=(p_map_wall)*data;
			map_delete_element_list(MAP_BUTTON, &wall->buttons);
			map_delete_item_list(&wall->items);
			break;
		case MAP_DOOR:
			door=(p_map_door)*data;
			map_delete_element_list(MAP_BUTTON, &door->buttons);
			map_delete_item_list(&door->itemsNE);
			map_delete_item_list(&door->itemsNW);
			map_delete_item_list(&door->itemsSE);
			map_delete_item_list(&door->itemsSW);
			break;
		case MAP_STAIRS:
			stairs=(p_map_stairs)*data;
			break;
		case MAP_TRAP:
			trap=(p_map_trap)*data;
			break;
		case MAP_PIT:
			pit=(p_map_pit)*data;
			break;
		case MAP_TELEPORTATION:
			teleportation=(p_map_teleportation)*data;
			break;
		case MAP_BUTTON:
			button=(p_map_button)*data;
			break;
		default:
			sprintf(tmp, "unexpected element %d to delete", type);
			error("game.map", tmp);
	}
	xfree((ptr)data);
}

void map_delete_element_list(const map_element type, p_list* element_list)
/* free an element list (list structure, cells and content) and resets its pointer */
{
	p_list	thelist;
	p_cell	c;

	if (element_list==NULL)
	{
		error("game.map", "attempting to delete a NULL element list pointer");
		return;
	}
		
	if (*element_list==NULL)
	{
		error("game.map", "attempting to delete a NULL element list structure");
		return;
	}

	thelist=*element_list;
	/* free the cells' content */
	c=thelist->head;
	while (c)
	{
		if (c->content)
			map_delete_element(type, (ptr*)&c->content);
		c=c->next;
	}

	/* remove cells */
	list_clear(thelist);

	/* free the list structure */
	xfree((ptr)element_list);
}

void map_delete_character(p_map_character* data)
{
	p_map_character thecharacter;

	if (data==NULL)
	{
		error("game.map", "attempting to delete a NULL character");
		return;
	}

	if (*data==NULL)
	{
		error("game.map", "attempting to delete a NULL character structure");
		return;
	}

	thecharacter=*data;
	xfree((ptr)&thecharacter->name);
	map_delete_item_list(&thecharacter->items);
	/* TODO: complete */

	xfree((ptr)*data);
}

void map_delete_character_list(p_list* character_list)
{
	p_list	thelist;
	p_cell	c;

	if (character_list==NULL)
	{
		error("game.map", "attempting to delete a NULL character list pointer");
		return;
	}
		
	if (*character_list==NULL)
	{
		error("game.map", "attempting to delete a NULL character list structure");
		return;
	}

	thelist=*character_list;
	/* free the cells' content */
	c=thelist->head;
	while (c)
	{
		if (c->content)
			map_delete_character((p_map_character*)&c->content);
		c=c->next;
	}

	/* remove cells */
	list_clear(thelist);

	/* free the list structure */
	xfree((ptr)character_list);
}

void map_delete_item(p_map_item* data)
{
	p_map_item theitem;

	if (data==NULL)
	{
		error("game.map", "attempting to delete a NULL item pointer");
		return;
	}

	if (*data==NULL)
	{
		error("game.map", "attempting to delete a NULL item structure");
		return;
	}

	theitem=*data;
	xfree((ptr)&theitem->name);
	map_delete_item_list(&theitem->items);
	/* TODO: complete */

	xfree((ptr)*data);
}

void map_delete_item_list(p_list* item_list)
{
	p_list	thelist;
	p_cell	c;

	if (item_list==NULL)
	{
		error("game.map", "attempting to delete a NULL item list pointer");
		return;
	}
		
	if (*item_list==NULL)
	{
		error("game.map", "attempting to delete a NULL item list structure");
		return;
	}

	thelist=*item_list;
	/* free the cells' content */
	c=thelist->head;
	while (c)
	{
		if (c->content)
			map_delete_item((p_map_item*)&c->content);
		c=c->next;
	}

	/* remove cells */
	list_clear(thelist);

	/* free the list structure */
	xfree((ptr)item_list);
}

#ifdef DEBUG
void debug_show_door_structure(p_map_door door)
{
	printf(	"door (%p):\n", door);
	if (door)
		printf(	"door (%p):\n"
				"\tUID=%lu\n"
				"\ttype=%d\n"
				"\torientation=%d\n"
				"\tstage=%d\n"
				"\tenabled=%d\n"
				"\tbuttons=%p\n"
				"\titemsNE=%p\n"
				"\titemsNW=%p\n"
				"\titemsSE=%p\n"
				"\titemsSW=%p\n",
					door,
					door->UID,
					door->type,	
					door->orientation,
					door->stage,
					door->enabled,
					door->buttons,
					door->itemsNE,
					door->itemsNW,
					door->itemsSE,
					door->itemsSW);
}
#endif


/****************************************************************************/
/*							'IN THE MAZE' MANAGEMENT						*/
/****************************************************************************/


str facing(character_facing direction)
{
	uint	i;
	char	tmp[STR_MAX];

	i=0;
	while (directions_table[i])
	{
		if (i==direction)
			return(directions_table[i]);
		i++;
	}
	sprintf(tmp, "unexpected direction %d", direction);
	fatal("game", tmp);
	return(NULL);
}

bool move_character(int new_x, int new_y, character_motion motion)
{
	map_element		tile;
	ptr				element;
	p_map_pit		pit;
	p_map_stairs	stairs;
	p_map_door		door;
	p_map_wall		wall;
/*	p_map_ground	ground;*/
	bool			collision;
	char			tmp[STR_MAX];

	collision=false;
	if (	(new_x>=0)
		AND (new_x<team.map->xsize)
		AND (new_y>=0)
		AND (new_y<team.map->ysize))
	{	
		map_get_tile(new_x, new_y, &tile, &element);
		switch (tile)
		{
			case MAP_GROUND:	/* free place: move to it! */
				/* collision=is there any character on this map tile?; */
				collision=false;

				if (NOT(collision))
				{
					team.position.x=new_x;
					team.position.y=new_y;
					team.position.motion=motion;
				}
				break;

			case MAP_PIT:	/* pit (hole) */
				pit=(p_map_pit)element;	/* mandatory */
				if (pit==NULL)
					fatal("game.maze", "undefined pit element in map");
				team.position.level=pit->level;
				team.position.x=pit->x;
				team.position.y=pit->y;
				if (pit->orientation!=CFACING_NEUTRAL)
					team.position.facing=pit->orientation;
				team.map=maps_table[team.position.level];
				break;
				
			case MAP_STAIRS:	/* stairs */
				stairs=(p_map_stairs)element;	/* mandatory */
				if (stairs==NULL)
					fatal("game.maze", "undefined stairs element in map");
				team.position.level=stairs->level;
				team.position.x=stairs->x;
				team.position.y=stairs->y;
				if (stairs->orientation==CFACING_NEUTRAL)
					team.position.facing=(team.position.facing+2)%4;
				else
					team.position.facing=stairs->orientation;
				team.map=maps_table[team.position.level];
				break;

			case MAP_DOOR:	/* doors */
				door=(p_map_door)element;	/* mandatory */
				if (door==NULL)
					fatal("game.maze", "undefined door element in map");
				if (door->stage>=DOOR_NEARLY_OPENED)
				{
					team.position.x=new_x;
					team.position.y=new_y;
					team.position.motion=motion;
				}
				else
					collision=true;
				break;

			case MAP_WALL:
				collision=true;
				break;

			case MAP_MAGICAL_WALL:
				wall=(p_map_wall)element;	/* mandatory */
				if (wall==NULL)
					fatal("game.maze", "undefined magical wall element in map");
				if (wall->enabled)
					collision=true;
				else
				{
					team.position.x=new_x;
					team.position.y=new_y;
					team.position.motion=motion;
				}
				break;

			default:	/* do nothing (don't move=collision) */
				sprintf(tmp, "unrecognized tile %d in map %d at (%d,%d)",
						tile, team.map->level, new_x, new_y);
				error("game.maze", tmp);
				collision=true;
		}
	}
	return(collision);
}

void maze_event_handler(SDL_Event* IO_event)
{
	character_motion	motion; /* character-relative direction */
	int					direction; /* geographical absolute direction */
	int					i;
	bool				collision;
	map_element			tile;
	ptr					element;
	p_map_door			door;
	p_map_button		button;
	int					base_x, base_y;

/*	if (get_current_game_handler()!=GHANDLER_MAZE)
		return;
*/
	motion=CMOTION_STANDING;
	if (IO_event)
		switch (IO_event->type)
		{
			case SDL_KEYDOWN:	/* keys */
				if (IO_keymod_pressed(SDLK_ESCAPE, KMOD_NONE))
				{
					gfx_clear(screen);
					gfx_flush();
					pop_game_handler();
					return;
				}
				else
				if (IO_keymod_pressed(SDLK_KP5, KMOD_NONE))
					motion=CMOTION_GOING_FRONT;
				else
				if (IO_keymod_pressed(SDLK_KP2, KMOD_NONE))
					motion=CMOTION_GOING_BACK;
				else
				if (IO_keymod_pressed(SDLK_KP1, KMOD_NONE))
					motion=CMOTION_GOING_LEFT;
				else
				if (IO_keymod_pressed(SDLK_KP3, KMOD_NONE))
					motion=CMOTION_GOING_RIGHT;
				else
				if (IO_keymod_pressed(SDLK_KP4, KMOD_NONE))
					motion=CMOTION_TURNING_LEFT;
				else
				if (IO_keymod_pressed(SDLK_KP6, KMOD_NONE))
					motion=CMOTION_TURNING_RIGHT;
/*				else
				if (IO_keymod_pressed(SDLK_DOWN, KMOD_LSHIFT))
					motion=CMOTION_U_TURNING;
*/				break;

			case SDL_MOUSEBUTTONDOWN:	/* mouse click */
				if (IO_mouse_clicked(SDL_BUTTON_LEFT))
				{
					/* TODO: manage left click */
					/* redraw=true; */

					base_x=GFX_MAZE_BASE_X;
					base_y=GFX_MAZE_BASE_Y;
					map_get_tile_relative(0, -1, &tile, &element);
					switch (tile)
					{
						case MAP_MAGICAL_WALL:
						case MAP_WALL:
							break;
						case MAP_DOOR:
							/* TODO support multiple buttons */
							door=(p_map_door)element;
							/* door has button? */
							if (door AND door->buttons)
							{
								if ((p_list)(door->buttons)==NULL)
									break;
								if ((p_cell)(p_list)(door->buttons)->head==NULL)
									break;
								button=(p_map_button)((p_cell)(p_list)(door->buttons)->head)->content;
								/* button is enabled? */
								if (button AND button->enabled)
								{
									/* button has been clicked? */
									if (IO_mouse_click_in_box(base_x+156+12, base_y+14+26, 8, 9))
									{
										/* change door's motion */
										switch (door->motion)
										{
											case DMOTION_NONE:
												if (door->stage==DOOR_OPENED)
													door->motion=DMOTION_CLOSING;
												else
													door->motion=DMOTION_OPENING;
												break;
											case DMOTION_OPENING:
												door->motion=DMOTION_CLOSING;
												break;
											case DMOTION_CLOSING:
												door->motion=DMOTION_OPENING;
												break;
										}
										timers_register_new(300, false, 0, TIMER_DOOR, (ptr)door, true);
									}
								}
							}
							break;
						case MAP_PIT:
							break;
						case MAP_TRAP:
							break;
						case MAP_STAIRS:
							break;
						default:
							NOP
					}

					base_x=GFX_KEYPAD_BASE_X;
					base_y=GFX_KEYPAD_BASE_Y;
					if (IO_mouse_click_in_box(base_x, base_y, 29, 22))
						motion=CMOTION_TURNING_LEFT;
					else
					if (IO_mouse_click_in_box(base_x+29, base_y, 29, 22))
						motion=CMOTION_GOING_FRONT;
					else
					if (IO_mouse_click_in_box(base_x+29+29, base_y, 29, 22))
						motion=CMOTION_TURNING_RIGHT;
					else
					if (IO_mouse_click_in_box(base_x, base_y+22, 29, 23))
						motion=CMOTION_GOING_LEFT;
					else
					if (IO_mouse_click_in_box(base_x+29, base_y+22, 29, 23))
						motion=CMOTION_GOING_BACK;
					else
					if (IO_mouse_click_in_box(base_x+29+29, base_y+22, 29, 23))
						motion=CMOTION_GOING_RIGHT;
				}
				else
				if (IO_mouse_clicked(SDL_BUTTON_RIGHT))
				{
					/* TODO: manage right click */
					/* redraw=true; */

					push_game_handler(GHANDLER_INVENTORY);
				}
				break;

			case SDL_MOUSEMOTION:	/* mouse motion */
				if (IO_mouse_has_moved())
				{
/*printf("(%d, %d)\n", mouse.x/video_conf.multiplier, mouse.y/video_conf.multiplier);*/
					/* TODO: set mouse pointer according to its position */
					IO_mouse_set_pointer(NULL);

					/* TODO: activate background parts according to the mouse position */

					/*redraw=true;*/
				}
				break;

			default:
				NOP
		}
		
	/* converts a character-relative motion to an absolute direction */
	direction=-1;
	collision=false;
	switch (motion)
	{
		case CMOTION_GOING_FRONT:
		case CMOTION_GOING_RIGHT:
		case CMOTION_GOING_BACK:
		case CMOTION_GOING_LEFT:
			i=0;
			while (movements_table[i].motion)
			{
				if (	(movements_table[i].motion==motion)
					AND (movements_table[i].facing==team.position.facing))
				{
					direction=(int)movements_table[i].movement;
					break;
				}
				i++;
			}
			/* motion*12+current_pos.facing*3+2*/
			break;
		case CMOTION_TURNING_LEFT:
			team.position.facing=abs((team.position.facing-1)%DIRECTIONS);
			break;
		case CMOTION_TURNING_RIGHT:
			team.position.facing=(team.position.facing+1)%DIRECTIONS;
			break;
		case CMOTION_U_TURNING:
			team.position.facing=(team.position.facing+2)%DIRECTIONS;
			break;
		default:
			NOP
	}

	/* move character */
	if (direction!=-1)
	{
		int	new_x, new_y;

		new_x=team.position.x;
		new_y=team.position.y;
		switch (direction)
		{
			case CMOVE_GOING_NORTH:
				new_y-=1;
				break;
			case CMOVE_GOING_EAST:
				new_x+=1;
				break;
			case CMOVE_GOING_SOUTH:
				new_y+=1;
				break;
			case CMOVE_GOING_WEST:
				new_x-=1;
				break;
			default:
				NOP
		}
		collision=move_character(new_x, new_y, motion);
	}

	/* force display when has been done or when no move is done (just to
		update the keypad drawing */
	keypad_old=keypad_set;
	keypad_set=motion;
	if (	(motion!=CMOTION_STANDING)
		OR	(keypad_set!=CMOTION_STANDING)
		OR	(keypad_old!=keypad_set))
	{
		keypad_set=motion;
		redraw=true;
		if (	(motion!=CMOTION_STANDING)
			AND	(collision==false))
			draw_alternate=NOT(draw_alternate);
	}
		
#ifdef DEBUG
	if (collision)
		info("game", "collision!");
#endif
}

void maze_init(void)
{
	info("game.maze", "starting module init..");

	IO_flush_events();

	info("game.maze", "module init completed");
}

void maze_reset(void)
{
	info("game.maze", "starting module reset..");

	info("game.maze", "module reset completed");
	pop_game_handler();
}


/****************************************************************************/
/*								TEAM MANAGEMENT								*/
/****************************************************************************/


p_team_character team_character_pointer(const ulong character_index)
{
	p_team_character	pointer;
	char				tmp[STR_MAX];
	
	pointer=team.red;
	switch (character_index)
	{
		case 0:
			pointer=team.red;
			break;
		case 1:
			pointer=team.yellow;
			break;
		case 2:
			pointer=team.green;
			break;
		case 3:
			pointer=team.blue;
			break;
		default:
			sprintf(tmp, "illegal team character index #%lu", character_index);
			error("game.team", tmp);
	}
	return(pointer);
}

ulong team_character_index(const p_team_character character_pointer)
{
	ulong	index;
	char	tmp[STR_MAX];
	
	index=0;
	if (character_pointer==team.red)
		index=0;
	else
	if (character_pointer==team.yellow)
		index=1;
	else
	if (character_pointer==team.green)
		index=2;
	else
	if (character_pointer==team.blue)
		index=3;
	else
	{
		sprintf(tmp, "illegal team character pointer %p", character_pointer);
		error("game.team", tmp);
	}
	return(index);
}

void team_load(const str datafile)
{
	fileHandle	in;
	char		tmp[STR_MAX];
	size_t		ret;
	ulong		cnt;

	info("game.team", "loading team..");

	in=fopen(datafile, "rb");
	if ((in==NULL) OR (datafile==NULL))
	{
		sprintf(tmp, "cannot load team from file %s", datafile);
		error("game.team", tmp);
	}

	/* load the characters */
	team_load_character(in, &team.red);
	team_load_character(in, &team.yellow);
	team_load_character(in, &team.green);
	team_load_character(in, &team.blue);

	/* other attributes */
	fload(&team.position, sizeof(struct coordinates));

	fload(&cnt, sizeof(cnt));
	team.map=maps_table[cnt];
	
	fload(&cnt, sizeof(cnt));
	team.leader=team_character_pointer(cnt);

	fload(&cnt, sizeof(cnt));
	team.front_left=team_character_pointer(cnt);
	fload(&cnt, sizeof(cnt));
	team.front_right=team_character_pointer(cnt);
	fload(&cnt, sizeof(cnt));
	team.rear_left=team_character_pointer(cnt);
	fload(&cnt, sizeof(cnt));
	team.rear_right=team_character_pointer(cnt);

	fclose(in);

	info("game.team", "team loaded");
}

void team_load_character(const fileHandle in, p_team_character* data)
{
}

void team_save(const str datafile)
{
	fileHandle	out;
	char		tmp[STR_MAX];
	size_t		ret;
	ulong		cnt;

	info("game.team", "saving team..");

	out=fopen(datafile, "wb");
	if ((out==NULL) OR (datafile==NULL))
	{
		sprintf(tmp, "cannot save team to file %s", datafile);
		error("game.team", tmp);
	}

	/* load the characters */
	team_save_character(out, team.red);
	team_save_character(out, team.yellow);
	team_save_character(out, team.green);
	team_save_character(out, team.blue);

	/* other attributes */
	fsave(&team.position, sizeof(struct coordinates));

	fsave(&cnt, sizeof(cnt));
	team.map=maps_table[cnt];
	
	cnt=team_character_index(team.leader);
	fsave(&cnt, sizeof(cnt));

	cnt=team_character_index(team.front_left);
	fsave(&cnt, sizeof(cnt));
	cnt=team_character_index(team.front_right);
	fsave(&cnt, sizeof(cnt));
	cnt=team_character_index(team.rear_left);
	fsave(&cnt, sizeof(cnt));
	cnt=team_character_index(team.rear_right);
	fsave(&cnt, sizeof(cnt));

	fclose(out);

	info("game.team", "team saved");
}

void team_save_character(const fileHandle out, p_team_character data)
{
}

void team_delete_character(p_team_character* data)
{
}

void team_init(void)
{
	info("game.team", "starting team init..");

	/* TODO: characters creation */

	team.leader=team.red;
	team.front_left=team.red;
	team.front_right=team.yellow;
	team.rear_left=team.green;
	team.rear_right=team.blue;
	team.position.motion=CMOTION_STANDING;

	info("game.team", "team init completed");
}

void team_reset(void)
{
	info("game.team", "starting team reset..");

	team_delete_character(&team.red);
	team_delete_character(&team.yellow);
	team_delete_character(&team.green);
	team_delete_character(&team.blue);
	team.leader=NULL;
	team.front_left=NULL;
	team.front_right=NULL;
	team.rear_left=NULL;
	team.rear_right=NULL;
	team.position.motion=CMOTION_STANDING;

	info("game.team", "team reset completed");
}

void team_enter_map(int level)
{
	info("game.team", "loading map parameters..");

	team.map=maps_table[level];
	if (team.map==NULL)
		fatal("game.team", "cannot load NULL map");

	team.position.x=team.map->initial_x;
	team.position.y=team.map->initial_y;
	team.position.level=team.map->level;
	team.position.facing=team.map->initial_facing;

	info("game.team", "map parameters loaded");
}


/****************************************************************************/
/*						'THE INVENTORY' MANAGEMENT							*/
/****************************************************************************/


void inventory_event_handler(SDL_Event* IO_event)
{
/*	if (get_current_game_handler()!=GHANDLER_INVENTORY)
		return;
*/
	if (IO_event)
		switch (IO_event->type)
		{
			case SDL_KEYDOWN:	/* keys */
				if (IO_keymod_pressed(SDLK_ESCAPE, KMOD_NONE))
				{
					gfx_clear(screen);
					gfx_flush();
					pop_game_handler();
					return;
				}
				break;

			case SDL_MOUSEBUTTONDOWN:	/* mouse click */
				if (IO_mouse_clicked(SDL_BUTTON_LEFT))
				{
					/* TODO: manage left click */
					/* redraw=true; */
					NOP
				}
				else
				if (IO_mouse_clicked(SDL_BUTTON_RIGHT))
				{
					/* TODO: manage right click */
					/* redraw=true; */

					gfx_clear(screen);
					gfx_flush();
					pop_game_handler();
					return;
				}
				break;

			case SDL_MOUSEMOTION:	/* mouse motion */
				if (IO_mouse_has_moved())
				{
/*printf("(%d, %d)\n", mouse.x/video_conf.multiplier, mouse.y/video_conf.multiplier);*/
					/* TODO: set mouse pointer according to its position */
					IO_mouse_set_pointer(NULL);

					/* TODO: activate background parts according to the mouse position */

					/*redraw=true;*/
				}
				break;

			default:
				NOP
		}
}

void inventory_init(void)
{
	info("inventory", "starting module init..");

	IO_flush_events();

	info("inventory", "module init completed");
}

void inventory_reset(void)
{
	info("inventory", "starting module reset..");

	info("inventory", "module reset completed");
}


/****************************************************************************/
/*							GAME MANAGEMENT									*/
/****************************************************************************/


void game_init(void)
{
	info("game", "starting module init..");

	IO_flush_events();

#ifdef DEBUG
	{
		int map_index;

		map_load_default_maps();
		map_save(SAVESDIR "/level0.map", 0);
		map_save(SAVESDIR "/level1.map", 1);
		for (map_index=0; map_index<MAPS; map_index++)
			maps_table[map_index]=NULL;
	}
#endif
	map_load(DATADIR "/maps/level0.map", 0);
	map_load(DATADIR "/maps/level1.map", 1);
	/* set palette */
	gfx_load_palette(RSC_BOARD1A);

	info("game", "setting current map & coordinates..");
	/* init map & player's position */
	team_init();
	team_enter_map(0);

	keypad_set=CMOTION_STANDING;
	keypad_old=CMOTION_STANDING;
	
	info("game", "module init completed");

	push_game_handler(GHANDLER_MAZE);
}

void game_reset(void)
{
	int map_index;

	info("game", "starting module reset..");

	team_reset();
	team.map=NULL;

	for (map_index=0; map_index<MAPS; map_index++)
		map_delete(map_index);
	
	info("game", "module reset completed");
}

void game_load(const str datafile)
{
	info("game", "loading game..");

	/* TODO */
	
	info("game", "game loaded");
}

void game_save(const str datafile)
{
	info("game", "saving game..");

	/* TODO */
	
	info("game", "game saved");
}

