#ifndef _game_h_
#define _game_h_

#include "util.h"
#include "gfx.h"

#define MAPS 2
#define DIRECTIONS 4

typedef enum
{
	CMOTION_STANDING=0,
	CMOTION_GOING_FRONT,
	CMOTION_GOING_RIGHT,
	CMOTION_GOING_BACK,
	CMOTION_GOING_LEFT,
	CMOTION_TURNING_LEFT,
	CMOTION_TURNING_RIGHT,
	CMOTION_U_TURNING
} character_motion;

typedef enum
{
	CFACING_NORTH=0,
	CFACING_EAST,
	CFACING_SOUTH,
	CFACING_WEST,
	CFACING_NEUTRAL
} character_facing;

typedef enum
{
	CMOVE_GOING_NORTH=0,
	CMOVE_GOING_EAST,
	CMOVE_GOING_SOUTH,
	CMOVE_GOING_WEST
} character_movement;

typedef enum
{
	MAP_GROUND=0,
	MAP_WALL,
	MAP_MAGICAL_WALL,
	MAP_DOOR,
	MAP_STAIRS,
	MAP_TRAP,
	MAP_PIT,
	MAP_TELEPORTATION,
	MAP_BUTTON
} map_element;

typedef enum
{
	DOOR_CLOSED=0,
	DOOR_NEARLY_CLOSED,
	DOOR_HALF_CLOSED,
	DOOR_NEARLY_OPENED,
	DOOR_OPENED,
	DOOR_BROKEN
} door_stage;

typedef enum
{
	DMOTION_NONE=0,
	DMOTION_OPENING,
	DMOTION_CLOSING
} door_motion;
	
struct coordinates
{
	int					x;
	int					y;
	int					level;
	character_facing	facing;
	character_motion	motion;
};

struct movement
{
	character_motion	motion;
	character_facing	facing;
	character_movement	movement;
};

struct map
{
	str					name;			/* name of the map */
	int					level;			/* its level */
	int					xsize;			/* its x size (west-east) */
	int					ysize;			/* its y size (north-south) */
	int					initial_x;		/* the starting x when entering from nowhere */
	int					initial_y;		/* the starting y when entering from nowhere */
	character_facing	initial_facing; /* the starting facing when entering from nowhere */
	map_element*		tiles;			/* map tiles array */
	ptr*				elements;		/* tiles' detail (elements) array */
	p_list				characters;		/* characters list (NPC) */
};

struct map_item
{
	ulong				UID;			/* unique identifier */
	gfx_item			type;			/* item type */
	str					name;			/* item name */
	int					weight;			/* item weight */
	p_list				items;			/* item list */
	/* TODO: complete the item attributes list */
};

struct map_character
{
	ulong				UID;			/* unique identifier */
	gfx_item			type;			/* character type */
	str					name;			/* character name */
	p_list				items;			/* item list */
	/* TODO: complete the character attribute list */
};

struct map_ground
{
	ulong				UID;			/* unique identifier */
	gfx_item			decoration;		/* optional decoration */
	p_list				itemsNE;		/* item list at north-east*/
	p_list				itemsNW;		/* item list at north-west*/
	p_list				itemsSE;		/* item list at south-east*/
	p_list				itemsSW;		/* item list at south-east*/
};

struct map_wall
{
	ulong				UID;			/* unique identifier */
	gfx_item			type;			/* wall type */
	int					enabled;		/* active? */
	int					visible;		/* visible? */
	gfx_item			decoration;		/* optional decoration */
	p_list				buttons;		/* optional buttons */
	p_list				items;			/* item list */
};	

struct map_door
{
	ulong				UID;			/* unique identifier */
	gfx_item			type;			/* door type */
	character_facing	orientation;	/* north/south orientation? (either it's east/west, obviously) */
	door_stage			stage;			/* opened? */
	door_motion			motion;			/* current motion */	
	bool				enabled;		/* enabled? */
	p_list				buttons;		/* optional buttons */
	p_list				itemsNE;		/* item list at north-east*/
	p_list				itemsNW;		/* item list at north-west*/
	p_list				itemsSE;		/* item list at south-east*/
	p_list				itemsSW;		/* item list at south-east*/
};

struct map_stairs
{
	ulong				UID;			/* unique identifier */
	gfx_item			type;			/* stairs type */
	character_facing	orientation;	/* direction to face to when coming from the stairs */
	int					x;				/* leading to x.. */
	int					y;				/* leading to y.. */
	int					level;			/* leading to level.. */
};

struct map_trap
{
	ulong				UID;			/* unique identifier */
	gfx_item			type;			/* trap type */
	int					enabled;		/* active? */
	int					visible;		/* active? */
	/* TODO: link to an action */
};

struct map_pit
{
	ulong				UID;			/* unique identifier */
	gfx_item			type;			/* pit type */
	int					enabled;		/* active? */
	int					visible;		/* visible? */
	character_facing	orientation;	/* direction to face to when falling from the pit */
	int					x;				/* leading to x.. */
	int					y;				/* leading to y.. */
	int					level;			/* leading to level.. */
};

struct map_teleportation
{
	ulong				UID;			/* unique identifier */
	gfx_item			type;			/* teleportation type */
	int					enabled;		/* active? */
	int					visible;		/* visible? */
	character_facing	orientation;	/* direction to face to when being teleported */
	int					x;				/* leading to x.. */
	int					y;				/* leading to y.. */
	int					level;			/* leading to level.. */
};

struct map_button
{
	ulong				UID;			/* unique identifier */
	gfx_item			type;			/* button type */
	int					enabled;		/* active? */
	int					visible;		/* visible? */
	/* TODO: link to an action */
};

typedef struct map* p_map;
typedef struct map_item* p_map_item;
typedef struct map_character* p_map_character;
typedef struct map_ground* p_map_ground;
typedef struct map_wall* p_map_wall;
typedef struct map_door* p_map_door;
typedef struct map_stairs* p_map_stairs;
typedef struct map_trap* p_map_trap;
typedef struct map_pit* p_map_pit;
typedef struct map_teleportation* p_map_teleportation;
typedef struct map_button* p_map_button;

struct team_character
{
	str		name;
	/* TODO: complete the character attribute list */
};

typedef struct team_character* p_team_character;

struct team_set
{
	p_team_character	red;
	p_team_character	yellow;
	p_team_character	green;
	p_team_character	blue;
	struct coordinates	position;
	p_map 				map;
	p_team_character	leader;
	p_team_character	front_left;
	p_team_character	front_right;
	p_team_character	rear_left;
	p_team_character	rear_right;
};


#ifdef DEBUG
void	map_load_default_maps(void);
#endif
void	map_load(const str datafile, const int map_index);
void	map_load_element(const fileHandle in, const map_element type, ptr* data);
void	map_load_element_list(const fileHandle in, const map_element type, p_list* element_list);
void	map_load_character(const fileHandle in, p_map_character* data);
void	map_load_character_list(const fileHandle in, p_list* character_list);
void	map_load_item(const fileHandle in, p_map_item* data);
void	map_load_item_list(const fileHandle in, p_list* item_list);

void	map_save(const str datafile, const int map_index);
void	map_save_element(const fileHandle out, const map_element type, ptr data);
void	map_save_element_list(const fileHandle out, const map_element type, p_list element_list);
void	map_save_character(const fileHandle out, p_map_character data);
void	map_save_character_list(const fileHandle out, p_list character_list);
void	map_save_item(const fileHandle out, p_map_item data);
void	map_save_item_list(const fileHandle out, p_list item_list);

void	map_get_tile(int x, int y, map_element* tile, ptr* element);
void	map_get_tile_relative(int front_switch, int side_switch, map_element* tile, ptr* element);

void	map_delete(const int map_index);
void	map_delete_element(const map_element type, ptr* data);
void	map_delete_element_list(const map_element type, p_list* element_list);
void	map_delete_character(p_map_character* data);
void	map_delete_character_list(p_list* character_list);
void	map_delete_item(p_map_item* data);
void	map_delete_item_list(p_list* item_list);

#ifdef DEBUG
void	debug_show_door_structure(p_map_door door);
#endif
		
str		facing(character_facing direction);
bool	move_character(int new_x, int new_y, character_motion motion);
void	maze_event_handler(SDL_Event* IO_event);
void	maze_init(void);
void	maze_reset(void);

p_team_character	team_character_pointer(const ulong character_index);
ulong				team_character_index(const p_team_character character_pointer);
void				team_load(const str datafile);
void				team_load_character(const fileHandle in, p_team_character* data);
void				team_save(const str datafile);
void				team_save_character(const fileHandle out, p_team_character data);
void				team_delete_character(p_team_character* data);
void				team_init(void);
void				team_reset(void);
void				team_enter_map(int level);

void	inventory_event_handler(SDL_Event* IO_event);
void	inventory_init(void);
void	inventory_reset(void);

void	game_init(void);
void	game_reset(void);
void	game_load(const str datafile);
void	game_save(const str datafile);
#endif
