#include "gfx.h"
#include "IO.h"
#include "game.h"
#include "core.h"

/*#define BENCHMARKS */

extern struct team_set			team;
extern struct mouse_status		mouse;
extern character_motion			keypad_set;
extern character_motion			keypad_old;


/****************************************************************************/
/*									TABLES									*/
/****************************************************************************/


struct gfx_data_set gfx_data_table[]=
{
	{	RSC_BOARD1A,	DATADIR "/gfx/set1A.png",	320,	240,	NULL,	false,	 0,	},
	{	RSC_BOARD2A,	DATADIR "/gfx/set2A.png",	320,	240,	NULL,	false,	 0	},
	{	RSC_BOARD1B,	DATADIR "/gfx/set1B.png",	320,	240,	NULL,	true,	10	},
	{	RSC_BOARD2B,	DATADIR "/gfx/set2B.png",	320,	240,	NULL,	true,	10	},
	{	RSC_BOARD1C,	DATADIR "/gfx/set1C.png",	320,	240,	NULL,	true,	10	},
	{	RSC_BOARD1D,	DATADIR "/gfx/set1D.png",	320,	240,	NULL,	false,	 0	},
	{	RSC_LOGO,		DATADIR "/gfx/logo.png",	320,	240,	NULL,	false,	 0	},
	{	RSC_TITLE,		DATADIR "/gfx/title.png",	320,	240,	NULL,	false,	 0	},
	{	RSC_MENU,		DATADIR "/gfx/menu.png",	320,	240,	NULL,	false,	 0	},
	{	RSC_CREDITS,	DATADIR "/gfx/credits.png",	320,	240,	NULL,	false,	 0	},
	{	RSC_UNKNOWN,	NULL,						  0,	  0,	NULL,	false,	 0	},
};

struct gfx_data_item gfx_items_table[]=
{
	/*  ident             dataset     x       y   width  height   tgt_x   tgt_y    masked?  mask */
	{	ITEM_BACKGROUND,		0,	  0,	  0,	224,	136,	  0,	  0,	false,	NULL	},

	{	ITEM_LEFT_WALL_0,		1,	284,	  0,	 32,	136,	  0,	  0,	false,	NULL	},
	{	ITEM_RIGHT_WALL_0,		1,	253,	  0,	 32,	136,	  0,	  0,	false,	NULL	},

	{	ITEM_LEFT_WALL_1,		1,	218,	  0,	 28,	111,	  0,	  0,	false,	NULL	},
	{	ITEM_FRONT_WALL_1,		1,	 33,	  0,	160,	111,	  0,	  0,	false,	NULL	},
	{	ITEM_RIGHT_WALL_1,		1,	  0,	  0,	 28,	111,	  0,	  0,	false,	NULL	},

	{	ITEM_LEFT_WALL_2,		1,	  0,	110,	 16,	 72,	  0,	  0,	false,	NULL	},
	{	ITEM_FRONT_WALL_2,		1,	  0,	110,	104,	 72,	  0,	  0,	false,	NULL	},
	{	ITEM_RIGHT_WALL_2,		1,	  0,	110,	 16,	 72,	  0,	  0,	false,	NULL	},

	{	ITEM_LEFT_WALL_3,		1,	  0,	110,	 12,	 50,	  0,	  0,	false,	NULL	},
	{	ITEM_FRONT_WALL_3,		1,	  0,	110,	 90,	 50,	  0,	  0,	false,	NULL	},
	{	ITEM_RIGHT_WALL_3,		1,	  0,	110,	 12,	 50,	  0,	  0,	false,	NULL	},

	{	ITEM_DOORWAY_1,			2,	  0,	  0,	 32,	123,	  0,	  0,	true,	NULL	},
	{	ITEM_LEFT_DOORWAY_2,	2,	 33,	  0,	 25,	 94,	  0,	  0,	true,	NULL	},
	{	ITEM_RIGHT_DOORWAY_2,	2,	 59,	  0,	 35,	 94,	  0,	  0,	true,	NULL	},
	{	ITEM_LEFT_DOORWAY_3,	2,	107,	  0,	 21,	 65,	  0,	  0,	true,	NULL	},
	{	ITEM_RIGHT_DOORWAY_3,	2,	 85,	  0,	 21,	 65,	  0,	  0,	true,	NULL	},

	{	ITEM_TRAP_1,			2,	  0,	172,	 60,	 20,	  0,	  0,	true,	NULL	},
	{	ITEM_TRAP_2,			2,	 33,	112,	 41,	 11,	  0,	  0,	true,	NULL	},

	{	ITEM_PIT_0,				3,	  0,	  0,	  0,	  0,	  0,	  0,	false,	NULL	},
	{	ITEM_PIT_1,				3,	  0,	  0,	  0,	  0,	  0,	  0,	false,	NULL	},
	{	ITEM_PIT_2,				3,	  0,	  0,	  0,	  0,	  0,	  0,	false,	NULL	},

	{	0,						0,	0,	0,	0,	0,	0,	0,	false,	NULL	}
};

bool						redraw;
bool						first_time;
bool						draw_alternate;
SDL_Surface*				screen;
SDL_Surface*				shadow;
struct display_set			display;
struct video_options_set	video_options;
struct FPS_manager			FPS_sync;


/****************************************************************************/
/*							GENERAL GFX MANAGEMENT							*/
/****************************************************************************/


void gfx_init(void)
{
	SDL_Surface*	tmp_screen;
	SDL_Rect**		modes;
	char			tmp[STR_MAX];
	int				i, ret;
	Uint32			transparency_key;

	info("video", "starting module init..");
	
	info("video", "setting video attributes..");
	sprintf(tmp, "%dx%dx%d [%x]", video_options.width, video_options.height,
									video_options.depth, video_options.SDL_flags);
	info("video", tmp);
	
	/* init video mode */
	modes=SDL_ListModes(NULL, video_options.SDL_flags);
	if (modes==(SDL_Rect**)0)
	{
		sprintf(tmp, "no video mode available (%s)", SDL_GetError());
		fatal("video", tmp);
	}
	else
		if (modes!=(SDL_Rect**)-1)
		{
			warning("video",
				"unable to set video mode, the best overall will be chosen if possible");
			info("video", "available modes:");
			i=0;
			while (modes[i])
			{
		    	sprintf(tmp, "%dx%d", modes[i]->w, modes[i]->h);
				info("video", tmp);
				i++;
			}
		}

	screen=SDL_SetVideoMode(video_options.width*video_options.scale,
							video_options.height*video_options.scale,
							video_options.depth, video_options.SDL_flags);
	if (!screen)
		fatal("video", "SDL video init error\n");

	if (screen->format->BitsPerPixel!=video_options.depth)
	{
		sprintf(tmp, "cannot init %d-bit video mode (%s)",
					video_options.depth, SDL_GetError());
		warning("video", tmp);
	}

	tmp_screen=SDL_CreateRGBSurface(video_options.SDL_flags,
									video_options.width*video_options.scale,
									video_options.height*video_options.scale,
									video_options.depth, 0, 0, 0, 0);
	if (tmp_screen==NULL)
	{
		sprintf(tmp, "cannot init RGB shadowed buffer (%s)", SDL_GetError());
		fatal("video", tmp);
	}
	shadow=SDL_DisplayFormat(tmp_screen);
	if (shadow==NULL)
	{
		sprintf(tmp, "cannot init RGB shadowed buffer (%s)", SDL_GetError());
		fatal("video", tmp);
	}
	SDL_FreeSurface(tmp_screen);
	
	info("video", "allocating buffers..");
	/* load gfx resources (sprites, fonts, etc.) */
	i=0;
	while (gfx_data_table[i].resource!=RSC_UNKNOWN)
	{
/*		gfx_data_table[i].buffer=SDL_LoadBMP(gfx_data_table[i].datafile); */
		gfx_data_table[i].buffer=IMG_Load(gfx_data_table[i].datafile);
		if (gfx_data_table[i].buffer==NULL)
		{
			sprintf(tmp, "cannot load resource #%d from '%s' (%s)",
						i, gfx_data_table[i].datafile, SDL_GetError());
			fatal("video", tmp);
		}

		sprintf(tmp, "resource #%d loaded in %p from '%s'",
				i, gfx_data_table[i].buffer, gfx_data_table[i].datafile);
		info("video", tmp);
		i++;
	}

	info("video", "configuring boards..");
	/* init board transparency & other SDL stuff */
	i=0;
	while (gfx_data_table[i].resource!=RSC_UNKNOWN)
	{
		tmp_screen=gfx_data_table[i].buffer;

		/* clears the color key */
		ret=SDL_SetColorKey(tmp_screen, 0, 0);
		if (ret==-1)
		{
			sprintf(tmp, "cannot reset transparency for resource #%d (%s)",
						i, SDL_GetError());
			fatal("video", tmp);
		}

		/* enable color key and RLE only for transparent boards */
		if (gfx_data_table[i].transparent)
		{
			if (gfx_data_table[i].transparency_key)
				transparency_key=gfx_data_table[i].transparency_key;
			else
				transparency_key=tmp_screen->format->colorkey;
			ret=SDL_SetColorKey(tmp_screen, SDL_SRCCOLORKEY|SDL_RLEACCEL, transparency_key);
			if (ret==-1)
			{
				sprintf(tmp, "cannot set transparency ($%x) for resource #%d (%s)",
							transparency_key, i, SDL_GetError());
				fatal("video", tmp);
			}
			else
			{
				sprintf(tmp, "transparency (%x) set for resource #%d",
							transparency_key, i);
				info("video", tmp);
			}
		}

		ret=SDL_SetAlpha(tmp_screen, 0, 0);
		if (ret==-1)
		{
			sprintf(tmp, "cannot reset alpha blending for resource #%d (%s)",
						i, SDL_GetError());
			fatal("video", tmp);
		}

		sprintf(tmp, "resource #%d configured", i);
		info("video", tmp);
		i++;
	}

	gfx_clear(screen);
	gfx_flush();
	redraw=false;
	first_time=true;
	draw_alternate=false;
	info("video", "module init completed");
}

void gfx_reset(void)
{
	int	i;

	info("video", "starting module reset..");

	info("video", "de-allocating buffers..");
	if (screen)
	{
		SDL_FreeSurface(screen);
		screen=NULL;
	}
	if (shadow)
	{
		SDL_FreeSurface(shadow);
		shadow=NULL;
	}

	i=0;
	while (gfx_data_table[i].resource!=RSC_UNKNOWN)
	{
		if (gfx_data_table[i].buffer)
		{
			SDL_FreeSurface(gfx_data_table[i].buffer);
			gfx_data_table[i].buffer=NULL;
		}
		i++;
	}
	info("video", "module reset completed");
}

void gfx_save_screenshot(void)
{
	char	tmp[STR_MAX];
	char	filename[STR_MAX];
	ulong	i;

	i=0;
	sprintf(filename, "%s/screenshot%lu.bmp", SAVESDIR, i);
	while (file_exists(filename))
	{
		i++;
		if (i==0)
		{
			error("video", "cannot save screenshot");
			return;
		}
		sprintf(filename, "%s/screenshot%lu.bmp", SAVESDIR, i);
	}

	if (SDL_SaveBMP(screen, filename)<0)
	{
		sprintf(tmp, "failed to save screenshot (%s)", filename);
		error("video", tmp);
	}
	else
	{
		sprintf(tmp, "screenshot saved (%s)", filename);
		info("video", tmp);
	}
}

byte gfx_get_pixel(SDL_Surface* surface, const int x, const int y)
{
	byte* peek=(byte*)((ulong)(surface->pixels)
					+y*surface->w*surface->format->BytesPerPixel
					+x*surface->format->BytesPerPixel);
	return(*peek);
}
		
void gfx_set_pixel(SDL_Surface* surface, const int x, const int y, const byte color)
{
	byte* poke;
	poke=(byte*)((ulong)(surface->pixels)
					+y*surface->w*surface->format->BytesPerPixel
					+x*surface->format->BytesPerPixel);
	*poke=color;
}

bool gfx_fade_in(int milliseconds)
/* screen must be ready to be flushed to hardware video */
{
	SDL_Palette*	palette;
	SDL_Color*		orig_colors;
	SDL_Color*		fade_colors;
	int				ncolors, c;
	float			alpha, alpha_inc;
	Uint32			start, stop, ticks;
	bool			ret;

	palette=screen->format->palette;
	if (palette==NULL)
	{
		warning("video", "non-palettized video mode, unable to use fading functions");
		gfx_flush();
		SDL_Delay(milliseconds);
		return(false);
	}
	ncolors=palette->ncolors;
	orig_colors=(SDL_Color*)malloc(ncolors*sizeof(SDL_Color));
	fade_colors=(SDL_Color*)malloc(ncolors*sizeof(SDL_Color));
	memcpy(orig_colors, palette->colors, ncolors*sizeof(SDL_Color));
	memcpy(fade_colors, orig_colors, ncolors*sizeof(SDL_Color));

	memset(fade_colors, 0, ncolors*sizeof(SDL_Color));
	SDL_SetColors(screen, fade_colors, 0, ncolors);
	gfx_flush();

	ret=false;
	alpha=0;
	alpha_inc=255.0/milliseconds;
	start=SDL_GetTicks();
	do
	{
		stop=SDL_GetTicks();
		ticks=stop-start;
		start=stop;
		if (ticks==0)
		{
			ticks=1;
			SDL_Delay(1);
		}
		alpha+=alpha_inc*ticks;

		if (alpha<=255)
		{
			for (c=0; c<ncolors; c++)
			{
				fade_colors[c].r=((float)orig_colors[c].r/255)*alpha;
				fade_colors[c].g=((float)orig_colors[c].g/255)*alpha;
				fade_colors[c].b=((float)orig_colors[c].b/255)*alpha;
			}
			SDL_SetColors(screen, fade_colors, 0, ncolors);
		}
		ret=IO_wait_until_event_or_time(1);
	}
	while ((alpha<255) AND (ret==false));
	if (ret==false)
		SDL_SetColors(screen, orig_colors, 0, ncolors);
	free(fade_colors);
	free(orig_colors);
	return(ret);
}

bool gfx_fade_out(int milliseconds)
/* screen must have been flushed to hardware video before */
{
	SDL_Palette*	palette;
	SDL_Color*		orig_colors;
	SDL_Color*		fade_colors;
	int				ncolors, c;
	float			alpha, alpha_inc;
	Uint32			start, stop, ticks;
	bool			ret;

	palette=screen->format->palette;
	if (palette==NULL)
	{
		warning("video", "non-palettized video mode, unable to use fading functions");
		gfx_flush();
		SDL_Delay(milliseconds);
		return(false);
	}
	ncolors=palette->ncolors;
	orig_colors=(SDL_Color*)malloc(ncolors*sizeof(SDL_Color));
	fade_colors=(SDL_Color*)malloc(ncolors*sizeof(SDL_Color));
	memcpy(orig_colors, palette->colors, ncolors*sizeof(SDL_Color));
	memcpy(fade_colors, orig_colors, ncolors*sizeof(SDL_Color));

	SDL_SetColors(screen, orig_colors, 0, ncolors);

	ret=false;
	alpha=255;
	alpha_inc=-255.0/milliseconds;
	start=SDL_GetTicks();
	do
	{
		stop=SDL_GetTicks();
		ticks=stop-start;
		start=stop;
		if (ticks==0)
		{
			ticks=1;
			SDL_Delay(1);
		}
		alpha+=alpha_inc*ticks;

		if (alpha>=0)
		{
			for (c=0; c<ncolors; c++)
			{
				fade_colors[c].r=((float)orig_colors[c].r/255)*alpha;
				fade_colors[c].g=((float)orig_colors[c].g/255)*alpha;
				fade_colors[c].b=((float)orig_colors[c].b/255)*alpha;
			}
			SDL_SetColors(screen, fade_colors, 0, ncolors);
		}
		ret=IO_wait_until_event_or_time(1);
	}
	while ((alpha>0) AND (ret==false));
	if (ret==false)
	{
		memset(fade_colors, 0, ncolors*sizeof(SDL_Color));
		SDL_SetColors(screen, fade_colors, 0, ncolors);
	}
	free(fade_colors);
	free(orig_colors);
	return(ret);
}

void gfx_blit(SDL_Surface* source, 
			const ulong src_x, const ulong src_y, const ulong src_w, const ulong src_h,
			SDL_Surface* target, const ulong tgt_x, const ulong tgt_y)
{
	SDL_Rect		src_rect, tgt_rect;
	int				x, y;
	uint			i, j;
/*	SDL_Color*		color;*/
	SDL_Palette*	palette;

	src_rect.x=src_x;
	src_rect.y=src_y;
	src_rect.w=src_w;
	src_rect.h=src_h;

	tgt_rect.x=tgt_x;
	tgt_rect.y=tgt_y;
	tgt_rect.w=src_w;
	tgt_rect.h=src_h;

	if ((source!=screen) OR (target!=shadow))
		SDL_BlitSurface(source, &src_rect, target, &tgt_rect);


	if (target==shadow)
	{
		Uint16	color_word;
		Uint32	color_long;
		ulong	base_offset, read_offset, write_offset;
		ulong	pixel_size, line_size;
#ifdef BENCHMARKS
		Uint32 ticks, frame, avg=0, max=get_current_game_handler()==GHANDLER_MAZE?100:1;
#endif

		base_offset=(ulong)(screen->pixels);
		pixel_size=screen->format->BytesPerPixel;
		line_size=screen->w*pixel_size;
		palette=screen->format->palette;

#ifdef BENCHMARKS
		for (frame=0; frame<max; frame++)
		{
		ticks=SDL_GetTicks();
#endif

		switch (video_options.scale)
		{
			case 4:
				/* TODO: interpolation (antialiasing) */
				/* optimized once: from 237->131 (average ticks) */
				for (y=video_options.height-1; y>=0; y--)
				{
					read_offset=base_offset+y*line_size;
					for (x=video_options.width-1; x>=0; x--)
					{
						color_long=*(byte*)(read_offset+x*pixel_size);
						color_long|=color_long<<8;
						color_long|=color_long<<16;
						i=x<<2;
						j=y<<2;
						write_offset=base_offset+j*line_size+i*pixel_size;
						*(Uint32*)(write_offset)=color_long;
						*(Uint32*)(write_offset+line_size)=color_long;
						*(Uint32*)(write_offset+line_size*2)=color_long;
						*(Uint32*)(write_offset+line_size*3)=color_long;
					}
				}
				SDL_UpdateRect(screen, tgt_x, tgt_y, tgt_rect.w<<2, tgt_rect.h<<2);
				break;
			case 3:
				/* TODO: interpolation (antialiasing) */
				/* optimized once: from 150->90 (average ticks) */
				for (y=video_options.height-1; y>=0; y--)
				{
					read_offset=base_offset+y*line_size;
					for (x=video_options.width-1; x>=0; x--)
					{
						color_word=*(byte*)(read_offset+x*pixel_size);
						color_word|=color_word<<8;
						i=(x<<1)+x;
						j=(y<<1)+y;
						write_offset=base_offset+j*line_size+i*pixel_size;
						*(Uint16*)(write_offset)=color_word;
						*(Uint16*)(write_offset+line_size)=color_word;
						*(Uint16*)(write_offset+line_size*2)=color_word;
						*(byte*)(write_offset+pixel_size*2)=(byte)color_word;
						*(byte*)(write_offset+line_size+pixel_size*2)=(byte)color_word;
						*(byte*)(write_offset+line_size*2+pixel_size*2)=(byte)color_word;
					}
				}
				SDL_UpdateRect(screen, tgt_x, tgt_y, (tgt_rect.w<<1)+tgt_rect.w, (tgt_rect.h<<1)+tgt_rect.h);
				break;
			case 2:
				/* TODO: interpolation */
				/* optimized once: from 75->44 (average ticks) */
				for (y=video_options.height-1; y>=0; y--)
				{
					read_offset=base_offset+y*line_size;
					for (x=video_options.width-1; x>=0; x--)
					{
						color_word=*(byte*)(read_offset+x*pixel_size);
/*						if (video_options.smooth_scaling)
						{
							color=&palette->colors[*(byte*)(read_offset+x*pixel_size+1)];
							color_word<<=8;
							color_word|=SDL_MapRGB(screen->format, color->r, color->g, color->b);
						}
						else
*/							color_word|=color_word<<8;
						i=x<<1;
						j=y<<1;
						write_offset=base_offset+j*line_size+i*pixel_size;
						*(Uint16*)(write_offset)=color_word;
						*(Uint16*)(write_offset+line_size)=color_word;
					}
				}				
				SDL_UpdateRect(screen, tgt_x, tgt_y, tgt_rect.w<<1, tgt_rect.h<<1);
				break;
			default:
				SDL_UpdateRect(screen, tgt_x, tgt_y, tgt_rect.w, tgt_rect.h);
		}

#ifdef BENCHMARKS
		ticks=SDL_GetTicks()-ticks;
		avg+=ticks;
		}
		printf("avg ticks = %f\n",(float)avg/(float)frame);
#endif		
	}
}

void gfx_clear(SDL_Surface* screen)
{
	SDL_FillRect(screen, NULL, 0);
}

void gfx_flush(void)
{
/*	gfx_blit(screen, 0, 0, video_options.width, video_options.height, shadow, 0, 0);*/
	gfx_blit(screen, 0, 0, 0, 0, shadow, 0, 0);
}

void gfx_set_display(SDL_Surface* screen/*, const fonts font, const font_sizes font_size,
					const int x, const int y, const int grid_w, const int grid_h*/)
{
	/* set display's surface */
	display.screen=screen;
	/* set active font */
/*	display.font=&font_table[font];
	display.font_size=font_size;
*/	/* set grid start coordonates */
/*	display.x=x;
	display.y=y;
*/	/* set grid values if non 0 */
/*	display.grid_w=grid_w?grid_w:display.font->w;
	display.grid_h=grid_h?grid_h:display.font->h;
*/}


/****************************************************************************/
/*							GFX RESOURCE MANAGEMENT							*/
/****************************************************************************/


SDL_Surface* gfx_get_resource_buffer(const gfx_resources resource)
{
	uint	i;
	char	tmp[STR_MAX];

	i=0;
	while (gfx_data_table[i].resource!=RSC_UNKNOWN)
	{
		if (gfx_data_table[i].resource==resource)
			return(gfx_data_table[i].buffer);
		i++;
	}
	sprintf(tmp, "cannot find resource's buffer #%d in table", resource);
	fatal("video", tmp);
	return(NULL);
}

int gfx_get_resource_width(const gfx_resources resource)
{
	uint	i;
	char	tmp[STR_MAX];

	i=0;
	while (gfx_data_table[i].resource!=RSC_UNKNOWN)
	{
		if (gfx_data_table[i].resource==resource)
			return(gfx_data_table[i].width);
		i++;
	}
	sprintf(tmp, "cannot find resource's width #%d in table", resource);
	fatal("video", tmp);
	return(0);
}

int gfx_get_resource_height(const gfx_resources resource)
{
	uint	i;
	char	tmp[STR_MAX];

	i=0;
	while (gfx_data_table[i].resource!=RSC_UNKNOWN)
	{
		if (gfx_data_table[i].resource==resource)
			return(gfx_data_table[i].height);
		i++;
	}
	sprintf(tmp, "cannot find resource's height #%d in table", resource);
	fatal("video", tmp);
	return(0);
}

void gfx_load_palette(const gfx_resources resource)
{
	SDL_Surface*	buffer;
	char			tmp[STR_MAX];
	SDL_Palette*	palette;

	buffer=gfx_get_resource_buffer(resource);
	if (buffer==NULL)
	{
		sprintf(tmp, "corrupted buffer pointer for resource #%d", resource);
		fatal("video", tmp);
	}
	palette=buffer->format->palette;
	if (palette)
		SDL_SetColors(screen, palette->colors, 0, palette->ncolors);
	else
		error("video", "can't get resource palette");
}


/****************************************************************************/
/*								GAME RENDERING								*/
/****************************************************************************/


bool draw_door(p_map_door door)
{
	character_facing orientation;

	if (door==NULL)
		return(false);

	orientation=door->orientation;
	switch (team.position.facing)
	{
		case CFACING_NORTH:
		case CFACING_SOUTH:
			if ((orientation==CFACING_NORTH) OR (orientation==CFACING_SOUTH))
				return(true);
			break;
		case CFACING_EAST:
		case CFACING_WEST:
			if ((orientation==CFACING_EAST) OR (orientation==CFACING_WEST))
				return(true);
			break;
		default:
			NOP
	}
	return(false);
}

void gfx_render_maze(void)
/* TODO: complete w/ missing gfx (stairs, doorway@sight+3) */
{
	SDL_Surface*	p0;
	SDL_Surface*	p1;
	SDL_Surface*	p2;
	map_element		tile, other_tile;
	ptr				element, other_element;
	p_map_door		door;
	p_map_wall		wall;
	p_map_button	button;
	int				base_x, base_y;

	if (draw_alternate)
	{	
		p0=gfx_get_resource_buffer(RSC_BOARD1A);
		p1=gfx_get_resource_buffer(RSC_BOARD1B);
	}
	else
	{
		p0=gfx_get_resource_buffer(RSC_BOARD2A);
		p1=gfx_get_resource_buffer(RSC_BOARD2B);
	}
	p2=gfx_get_resource_buffer(RSC_BOARD1C);

	base_x=GFX_MAZE_BASE_X;
	base_y=GFX_MAZE_BASE_Y;

	/* background (ceilar & ground at all sights) */

	gfx_blit(p0, 0, 0, 224, 136, screen, base_x+0, base_y+0);

	
	/******* sight +3 *******/

	/* left side */
	map_get_tile_relative(-1, -3, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 165, 111, 73, 51, screen, base_x+0, base_y+25);
			gfx_blit(p1, 240, 111, 12, 51, screen, base_x+73, base_y+25);
			break;
		case MAP_DOOR: /* TODO */
			door=(p_map_door)element;
			if (draw_door(door))
			{
				gfx_blit(p2, 94, 66, 12, 44, screen, base_x+0, base_y+27);		/* left */
/*				gfx_blit(p2, 107, 66, 12, 44, screen, base_x+52, base_y+27);*/	/* right */
				gfx_blit(p2, 119, 98, 48, 1, screen, base_x+8, base_y+29);		/* top */
				switch (door->stage)
				{
					case DOOR_CLOSED:
						gfx_blit(p2, 195, 62, 47, 39, screen, base_x+9, base_y+30);
						break;
					case DOOR_NEARLY_CLOSED:
						gfx_blit(p2, 195, 71, 47, 31, screen, base_x+9, base_y+29);
						break;
					case DOOR_HALF_CLOSED:
						gfx_blit(p2, 195, 80, 47, 22, screen, base_x+9, base_y+29);
						break;
					case DOOR_NEARLY_OPENED:
						gfx_blit(p2, 195, 89, 47, 13, screen, base_x+9, base_y+29);
						break;
					case DOOR_OPENED:
						break;
					case DOOR_BROKEN:
						/* TODO: use the correct sprite (need to be implemented */
						gfx_blit(p2, 195, 62, 47, 39, screen, base_x+9, base_y+30);
						break;
				}
			}
			break;
		case MAP_PIT: /* TODO */
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		default:
			NOP
	}

	/* right side */
	map_get_tile_relative(+1, -3, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 137, 111, 12, 51, screen, base_x+140, base_y+25);
			gfx_blit(p1, 148, 111, 72, 51, screen, base_x+152, base_y+25);
			break;
		case MAP_DOOR: /* TODO */
			door=(p_map_door)element;
			if (draw_door(door))
			{
/*				gfx_blit(p2, 94, 66, 12, 44, screen, base_x+140, base_y+27);*/		/* left */
				gfx_blit(p2, 107, 66, 12, 44, screen, base_x+204, base_y+27);	/* right */
				gfx_blit(p2, 119, 98, 48, 1, screen, base_x+152, base_y+29);		/* top */
				switch (door->stage)
				{
					case DOOR_CLOSED:
						gfx_blit(p2, 195, 62, 47, 39, screen, base_x+152, base_y+30);
						break;
					case DOOR_NEARLY_CLOSED:
						gfx_blit(p2, 195, 71, 47, 31, screen, base_x+152, base_y+29);
						break;
					case DOOR_HALF_CLOSED:
						gfx_blit(p2, 195, 80, 47, 22, screen, base_x+152, base_y+29);
						break;
					case DOOR_NEARLY_OPENED:
						gfx_blit(p2, 195, 89, 47, 13, screen, base_x+152, base_y+29);
						break;
					case DOOR_OPENED:
						break;
					case DOOR_BROKEN:
						/* TODO: use the correct sprite (need to be implemented */
						gfx_blit(p2, 195, 62, 47, 39, screen, base_x+152, base_y+30);
						break;
				}
			}
			break;
		case MAP_PIT: /* TODO */
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		default:
			NOP
	}

	/* ground */

	/* front */
	map_get_tile_relative(0, -3, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 156, 111, 76, 51, screen, base_x+74, base_y+25);
			break;
		case MAP_DOOR: /* TODO */
			door=(p_map_door)element;
			if (draw_door(door))
			{
				gfx_blit(p2, 94, 66, 12, 44, screen, base_x+80, base_y+27);		/* left */
				gfx_blit(p2, 107, 66, 12, 44, screen, base_x+132, base_y+27);	/* right */
				gfx_blit(p2, 119, 98, 48, 1, screen, base_x+88, base_y+29);		/* top */
				switch (door->stage)
				{
					case DOOR_CLOSED:
						gfx_blit(p2, 195, 62, 47, 39, screen, base_x+89, base_y+30);
						break;
					case DOOR_NEARLY_CLOSED:
						gfx_blit(p2, 195, 71, 47, 31, screen, base_x+89, base_y+29);
						break;
					case DOOR_HALF_CLOSED:
						gfx_blit(p2, 195, 80, 47, 22, screen, base_x+89, base_y+29);
						break;
					case DOOR_NEARLY_OPENED:
						gfx_blit(p2, 195, 89, 47, 13, screen, base_x+89, base_y+29);
						break;
					case DOOR_OPENED:
						break;
					case DOOR_BROKEN:
						/* TODO: use the correct sprite (need to be implemented */
						gfx_blit(p2, 195, 62, 47, 39, screen, base_x+89, base_y+30);
						break;
				}
			}
			break;
		case MAP_PIT: /* TODO */
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		default:
			NOP
	}

	/******* sight +2 *******/

	/* left side */
	map_get_tile_relative(-1, -2, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 61, 111, 75, 71, screen, base_x+0, base_y+20);
			break;
		case MAP_DOOR: /* TODO */
			door=(p_map_door)element;
			if (draw_door(door))
			{
   				gfx_blit(p2, 85, 0, 21, 65, screen, base_x+149, base_y+22);		/* left */
	/*   			gfx_blit(p2, 107, 0, 21, 65, screen, base_x+142, base_y+22);*/	/* right */
				gfx_blit(p2, 119, 100, 66, 2, screen, base_x+79, base_y+23);	/* top */
				switch (door->stage)
				{
					case DOOR_CLOSED:
						gfx_blit(p2, 195, 0, 52, 59, screen, base_x+149+18, base_y+25);
						break;
					case DOOR_NEARLY_CLOSED:
						gfx_blit(p2, 195, 15, 52, 46, screen, base_x+149+18, base_y+25);
						break;
					case DOOR_HALF_CLOSED:
						gfx_blit(p2, 195, 30, 52, 31, screen, base_x+149+18, base_y+25);
						break;
					case DOOR_NEARLY_OPENED:
						gfx_blit(p2, 195, 44, 52, 17, screen, base_x+149+18, base_y+25);
						break;
					case DOOR_OPENED:
						break;
					case DOOR_BROKEN:
						/* TODO: use the correct sprite (need to be implemented */
						gfx_blit(p2, 195, 0, 52, 60, screen, base_x+149+18, base_y+25);
						break;
				}
			}
			break;
		case MAP_PIT: /* TODO */
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		default:
			NOP
	}

	/* right side */
	map_get_tile_relative(+1, -2, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 0, 111, 75, 71, screen, base_x+149, base_y+20);
			break;
		case MAP_DOOR: /* TODO */
			door=(p_map_door)element;
			if (draw_door(door))
			{
   				gfx_blit(p2, 85, 0, 21, 65, screen, base_x+149, base_y+22);		/* left */
	/*   			gfx_blit(p2, 107, 0, 21, 65, screen, base_x+142, base_y+22);*/	/* right */
				gfx_blit(p2, 119, 100, 66, 2, screen, base_x+79, base_y+23);	/* top */
				switch (door->stage)
				{
					case DOOR_CLOSED:
						gfx_blit(p2, 195, 0, 52, 59, screen, base_x+149+18, base_y+25);
						break;
					case DOOR_NEARLY_CLOSED:
						gfx_blit(p2, 195, 15, 52, 46, screen, base_x+149+18, base_y+25);
						break;
					case DOOR_HALF_CLOSED:
						gfx_blit(p2, 195, 30, 52, 31, screen, base_x+149+18, base_y+25);
						break;
					case DOOR_NEARLY_OPENED:
						gfx_blit(p2, 195, 44, 52, 17, screen, base_x+149+18, base_y+25);
						break;
					case DOOR_OPENED:
						break;
					case DOOR_BROKEN:
						/* TODO: use the correct sprite (need to be implemented */
						gfx_blit(p2, 195, 0, 52, 60, screen, base_x+149+18, base_y+25);
						break;
				}
			}
			break;
		case MAP_PIT: /* TODO */
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		default:
			NOP
	}

	/* ground */

	/* front */
	map_get_tile_relative(0, -2, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 16, 111, 104, 71, screen, base_x+60, base_y+20);
			break;
		case MAP_DOOR:
   			gfx_blit(p2, 85, 0, 21, 65, screen, base_x+61, base_y+22);		/* left */
   			gfx_blit(p2, 107, 0, 21, 65, screen, base_x+142, base_y+22);	/* right */
			gfx_blit(p2, 119, 100, 66, 2, screen, base_x+79, base_y+23);	/* top */
			door=(p_map_door)element;
			switch (door->stage)
			{
				case DOOR_CLOSED:
					gfx_blit(p2, 195, 0, 64, 59, screen, base_x+80, base_y+25);
					break;
				case DOOR_NEARLY_CLOSED:
					gfx_blit(p2, 195, 15, 64, 46, screen, base_x+80, base_y+25);
					break;
				case DOOR_HALF_CLOSED:
					gfx_blit(p2, 195, 30, 64, 31, screen, base_x+80, base_y+25);
					break;
				case DOOR_NEARLY_OPENED:
					gfx_blit(p2, 195, 44, 64, 17, screen, base_x+80, base_y+25);
					break;
				case DOOR_OPENED:
					break;
				case DOOR_BROKEN:
					/* TODO: use the correct sprite (need to be implemented */
					gfx_blit(p2, 195, 0, 64, 60, screen, base_x+80, base_y+25);
					break;
			}
			break;
		case MAP_PIT:
			gfx_blit(p2, 33, 112, 41, 11, screen, base_x+90, base_y+79);
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		default:
			NOP
	}

	/******* sight +1 *******/

	/* left side */
	map_get_tile_relative(-1, -1, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 188, 0, 60, 111, screen, base_x+0, base_y+9);
			break;
		case MAP_DOOR:
			gfx_blit(p2, 113+25, 103, 32, 6, screen, base_x+0, base_y+14);	/* top */
			door=(p_map_door)element;
			switch (door->stage)
			{
				case DOOR_CLOSED:
					gfx_blit(p2, 92+96-32, 113, 32, 86, screen, base_x+0, base_y+17);
					break;
				case DOOR_NEARLY_CLOSED:
					gfx_blit(p2, 92+96-32, 135, 32, 65, screen, base_x+0, base_y+17);
					break;
				case DOOR_HALF_CLOSED:
					gfx_blit(p2, 92+96-32, 157, 32, 43, screen, base_x+0, base_y+17);
					break;
				case DOOR_NEARLY_OPENED:
					gfx_blit(p2, 92+96-32, 179, 32, 21, screen, base_x+0, base_y+17);
					break;
				case DOOR_OPENED:
					break;
				case DOOR_BROKEN:
					/* TODO: use the correct sprite (need to be implemented */
					gfx_blit(p2, 92+96-32, 113, 32, 85, screen, base_x+0, base_y+17);
					break;
			}
			break;
		case MAP_PIT: /* TODO */
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		default:
			NOP
	}

	/* right side */
	map_get_tile_relative(+1, -1, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 0, 0, 60, 111, screen, base_x+164, base_y+9);
			break;
		case MAP_PIT: /* TODO */
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		case MAP_DOOR:
			gfx_blit(p2, 113+25, 103, 32, 6, screen, base_x+192, base_y+14);	/* top */
			door=(p_map_door)element;
			switch (door->stage)
			{
				case DOOR_CLOSED:
					gfx_blit(p2, 92, 113, 32, 86, screen, base_x+192, base_y+17);
					break;
				case DOOR_NEARLY_CLOSED:
					gfx_blit(p2, 92, 135, 32, 65, screen, base_x+192, base_y+17);
					break;
				case DOOR_HALF_CLOSED:
					gfx_blit(p2, 92, 157, 32, 43, screen, base_x+192, base_y+17);
					break;
				case DOOR_NEARLY_OPENED:
					gfx_blit(p2, 92, 179, 32, 21, screen, base_x+192, base_y+17);
					break;
				case DOOR_OPENED:
					break;
				case DOOR_BROKEN:
					/* TODO: use the correct sprite (need to be implemented */
					gfx_blit(p2, 92, 113, 32, 85, screen, base_x+192, base_y+17);
					break;
			}
			break;
		default:
			NOP
	}

	/* ground */

	/* front */
	map_get_tile_relative(0, -1, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 44, 0, 160, 111, screen, base_x+32, base_y+9);
			map_get_tile_relative(0, 0, &other_tile, &other_element);
			if (other_tile==MAP_DOOR)
				gfx_blit(p2, 0, 0, 32, 123, screen, base_x+97, base_y+0);
			break;
		case MAP_DOOR:
			gfx_blit(p2, 59, 0, 25, 94, screen, base_x+43, base_y+14);	/* left */
			gfx_blit(p2, 33, 0, 25, 94, screen, base_x+156, base_y+14);	/* right */
			gfx_blit(p2, 119, 103, 104, 6, screen, base_x+60, base_y+14);	/* top */
			door=(p_map_door)element;
			switch (door->stage)
			{
				case DOOR_CLOSED:
					gfx_blit(p2, 92, 113, 96, 86, screen, base_x+64, base_y+17);
					break;
				case DOOR_NEARLY_CLOSED:
					gfx_blit(p2, 92, 135, 96, 65, screen, base_x+64, base_y+17);
					break;
				case DOOR_HALF_CLOSED:
					gfx_blit(p2, 92, 157, 96, 43, screen, base_x+64, base_y+17);
					break;
				case DOOR_NEARLY_OPENED:
					gfx_blit(p2, 92, 179, 96, 21, screen, base_x+64, base_y+17);
					break;
				case DOOR_OPENED:
					break;
				case DOOR_BROKEN:
					/* TODO: use the correct sprite (need to be implemented */
					gfx_blit(p2, 92, 113, 96, 85, screen, base_x+64, base_y+17);
					break;
			}
			/* TODO support multiple buttons */
			if (door->buttons)
			{
				button=(p_map_button)(door->buttons)->head;
				if (button AND button->visible)
					gfx_blit(p2, 119, 67, 8, 9, screen, base_x+156+12, base_y+14+26);
			}
			break;
		case MAP_PIT:
			gfx_blit(p2, 0, 172, 61, 20, screen, base_x+83, base_y+94);
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		default:
			NOP
	}

	/******* sight 0 *******/

	/* left side */
	map_get_tile_relative(-1, 0, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 284, 0, 32, 136, screen, base_x+0, base_y+0);
			break;
		case MAP_PIT: /* TODO */
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		default:
			NOP
	}	

	/* right side */
	map_get_tile_relative(+1, 0, &tile, &element);
	switch (tile)
	{
		case MAP_MAGICAL_WALL:
			wall=(p_map_wall)element;
			if (wall->visible==false)
				break;
		case MAP_WALL:
			gfx_blit(p1, 253, 0, 32, 136, screen, base_x+192, base_y+0);
			break;
		case MAP_PIT: /* TODO */
			break;
		case MAP_TRAP: /* TODO */
			break;
		case MAP_STAIRS: /* TODO */
			break;
		default:
			NOP
	}

	/* ground */
}

void gfx_render_keypad(void)
{
	SDL_Surface*	board;
	int				source_x, source_y;
	int				base_x, base_y;
	int				x_inc, y_inc;
	int				width;

	source_x=224;
	source_y=0;
	base_x=GFX_KEYPAD_BASE_X;
	base_y=GFX_KEYPAD_BASE_Y;	
	board=gfx_get_resource_buffer(RSC_BOARD1A);

	gfx_blit(board, source_x,source_y, 87, 45, screen, base_x, base_y);
	if (keypad_set!=keypad_old)
	{
		/* only check keypad_set since the UNSET whole pad is copied unconditionally */
		switch (keypad_set)
		{
			case CMOTION_TURNING_LEFT:
				x_inc=1;
				y_inc=1;
				width=28;
				break;
			case CMOTION_GOING_FRONT:
				x_inc=1+28+1;
				y_inc=1;
				width=27;
				break;
			case CMOTION_TURNING_RIGHT:
				x_inc=1+28+1+27+1;
				y_inc=1;
				width=28;
				break;
			case CMOTION_GOING_LEFT:
				x_inc=1;
				y_inc=1+21+1;
				width=28;
				break;
			case CMOTION_GOING_BACK:
				x_inc=1+28+1;
				y_inc=1+21+1;
				width=27;
				break;
			case CMOTION_GOING_RIGHT:
				x_inc=1+28+1+27+1;
				y_inc=1+21+1;
				width=28;
				break;
			default:
				return;
		}
		gfx_blit(board, source_x+x_inc, source_y+44+y_inc, width, 21,
				screen, base_x+x_inc, base_y+y_inc);
	}
}

void gfx_scene_display(void)
{
#ifdef DEBUG
	int				i, j;
	map_element		tile;
	ptr				element;
	byte			c;
	char			tmp[STR_MAX];
#endif
	
#ifdef DEBUG	
	printf("team: (%d, %d) facing %s, motion %d\n", team.position.x, team.position.y,
			facing(team.position.facing), team.position.motion);
	printf("map: #%d '%s' (%d, %d)\n", team.map->level, team.map->name, team.map->xsize, team.map->ysize);
	for (j=0; j<team.map->ysize; j++)
	{
		for (i=0; i<team.map->xsize; i++)
		{
			if ((i==team.position.x) AND (j==team.position.y))
				c='X';
			else
			{
				map_get_tile(i, j, &tile, &element);
				switch (tile)
				{
					case MAP_GROUND:
						c=' ';
						break;

					case MAP_PIT:
						c='T';
						break;

					case MAP_STAIRS:
						c='S';
						break;

					case MAP_DOOR:
						c='D';
						break;

					case MAP_MAGICAL_WALL:
						c='M';
						break;

					case MAP_WALL:
						c=128;
						break;

					default:
						sprintf(tmp, "unrecognized tile %d in map %d at (%d,%d)",
								tile, team.map->level, i, j);
						error("video", tmp);
						c='?';
				}
			}
			printf("%c", c);
		}
		printf("\n");
	}
#endif
	/* TODO: update other panels here */
	gfx_clear(screen);
	gfx_render_maze();
	gfx_render_keypad();
#ifdef DEBUG
	printf("redrawed SCENE\n");
#endif
}

void gfx_inventory_display(void)
{
#ifdef DEBUG
	printf("redrawed INVENTORY\n");
#endif
}

void gfx_menu_display(void)
{
	SDL_Surface*	data;
	gfx_resources	rsc;

	rsc=RSC_MENU;
	gfx_load_palette(rsc);
	data=gfx_get_resource_buffer(rsc);
	gfx_blit(data, 0, 0, gfx_get_resource_width(rsc),
						gfx_get_resource_height(rsc), screen, 0, 0);
#ifdef DEBUG
	printf("redrawed MENU\n");
#endif
}

void gfx_draw_screen(void)
{
	char	tmp[STR_MAX];
	bool	flush;

	flush=false;
	if (redraw OR first_time)
	{
	/*	gfx_sync(); */
		switch (get_current_game_handler())
		{
			case GHANDLER_MENU:
				gfx_clear(screen);
				gfx_menu_display();
				flush=true;
				break;
			case GHANDLER_MAZE:
				gfx_clear(screen);
				gfx_scene_display();
				flush=true;
				break;
			case GHANDLER_INVENTORY:
				gfx_clear(screen);
				gfx_inventory_display();
				flush=true;
				break;
			case GHANDLER_FILE_MANAGER:
				break;
			case GHANDLER_OPTIONS:
				break;
			case GHANDLER_HELP:
				break;
			case GHANDLER_PAUSED:
				break;
			default:
				sprintf(tmp, "illegal call to gfx_draw_screen: bad game handler %d",
						get_current_game_handler());
				fatal("video", tmp);
		}

		/* TODO display the mouse's cursor (pointer+position) */
		if (mouse.pointer==NULL)
			SDL_ShowCursor(SDL_ENABLE);

		if (flush)
			gfx_flush();
	}
	first_time=false;
	redraw=false;
}


/****************************************************************************/
/*							FRAMERATE MANAGEMENT							*/
/****************************************************************************/

/* these framerate management routs have been roughly taken from SDL_gfx */
void init_framerate(struct FPS_manager* manager)
{
    manager->framecount = 0;
    manager->rate = FPS_DEFAULT;
    manager->rateticks = (1000.0/(float)FPS_DEFAULT);
    manager->lastticks = SDL_GetTicks();
}

int set_framerate(struct FPS_manager* manager, int rate)
{
    if ((rate>=FPS_LOWER_LIMIT) AND (rate<=FPS_UPPER_LIMIT))
	{
		manager->framecount=0;
		manager->rate=rate;
		manager->rateticks=(1000.0/(float)rate);
		return(0);
    }
	else
		return(-1);
}

int get_framerate(struct FPS_manager* manager)
{
	return(manager->rate);
}

void framerate_delay(struct FPS_manager* manager)
{
    Uint32 current_ticks, target_ticks, the_delay;

    /* next frame */
    manager->framecount++;

    /* get/calc ticks */
    current_ticks=SDL_GetTicks();
    target_ticks=manager->lastticks+(Uint32)((float)manager->framecount*manager->rateticks);

    if (current_ticks<=target_ticks)
	{
		the_delay=target_ticks-current_ticks;
		SDL_Delay(the_delay);
    }
	else
	{
		manager->framecount=0;
		manager->lastticks=SDL_GetTicks();
    }
}


/****************************************************************************/
/*								BITMAPS FONTS								*/
/****************************************************************************/


