#ifndef _gfx_h_
#define _gfx_h_

#include "util.h"

#define GFX_MAZE_BASE_X 0
#define GFX_MAZE_BASE_Y 33
#define GFX_KEYPAD_BASE_X 232
#define GFX_KEYPAD_BASE_Y 124

typedef enum
{
	RSC_BOARD1A,
	RSC_BOARD2A,
	RSC_BOARD1B,
	RSC_BOARD2B,
	RSC_BOARD1C,
	RSC_BOARD1D,
	RSC_LOGO,
	RSC_TITLE,
	RSC_MENU,
	RSC_CREDITS,
	RSC_UNKNOWN
} gfx_resources;

typedef enum
{
	ITEM_BACKGROUND=1,	/* all-range ground & ceilar */

	ITEM_LEFT_WALL_0,	/* range */
	ITEM_RIGHT_WALL_0,

	ITEM_LEFT_WALL_1,
	ITEM_FRONT_WALL_1,
	ITEM_RIGHT_WALL_1,

	ITEM_LEFT_WALL_2,
	ITEM_FRONT_WALL_2,
	ITEM_RIGHT_WALL_2,

	ITEM_LEFT_WALL_3,
	ITEM_FRONT_WALL_3,
	ITEM_RIGHT_WALL_3,

	ITEM_DOORWAY_1,
	ITEM_LEFT_DOORWAY_2,
	ITEM_RIGHT_DOORWAY_2,
	ITEM_LEFT_DOORWAY_3,
	ITEM_RIGHT_DOORWAY_3,
	ITEM_WOODEN_DOOR,
	ITEM_IRON_DOOR,

	ITEM_STONE_WALL,
	ITEM_INVISIBLE_WALL,

	ITEM_BLUE_TELEPORTATION,

	ITEM_STAIRS_UP,
	ITEM_STAIRS_DOWN,
	
	ITEM_TRAP_1,
	ITEM_TRAP_2,

	ITEM_PIT_0,
	ITEM_PIT_1,
	ITEM_PIT_2,

	ITEM_BUTTON_0
} gfx_item;

struct gfx_data_set
{
	gfx_resources	resource;
	str				datafile;
	int				width;
	int				height;
	SDL_Surface*	buffer;
	bool			transparent;
	Uint32			transparency_key;
};

struct gfx_font_set
{
	str				datafile;
	int				width;
	int				height;
	SDL_Surface*	buffer;
};

struct gfx_data_item
{
	gfx_item	item;
	int			data;
	int			x;
	int			y;
	int			w;
	int			h;
	int			target_x;
	int			target_y;
	bool		masked;
	ptr			mask;
};

struct display_set
{
	SDL_Surface*	screen;
/*
	struct font_set	font;
	font_sizes		font_size;
	int				x;
	int				y;
	int				grid_w;
	int				grid_h;
*/
};

struct video_options_set
{
	int		width;
	int		height;
	int		depth;
	int		scale;
	bool	smooth_scaling;
	int		SDL_flags;
	int		FPS;
};

#define FPS_UPPER_LIMIT	200
#define FPS_LOWER_LIMIT	1
#define FPS_DEFAULT		30

struct FPS_manager
{
	Uint32	framecount;
	float	rateticks;
	Uint32	lastticks;
	Uint32	rate;
};


void			gfx_init(void);
void			gfx_reset(void);

void			gfx_save_screenshot(void);

byte			gfx_get_pixel(SDL_Surface* surface, const int x, const int y);
void			gfx_set_pixel(SDL_Surface* surface, const int x, const int y, const byte color);
bool			gfx_fade_in(int milliseconds);
bool			gfx_fade_out(int milliseconds);
void			gfx_alpha_blit(int alpha, SDL_Surface* source, 
					const ulong src_x, const ulong src_y, const ulong src_w, const ulong src_h,
					SDL_Surface* target, const ulong tgt_x, const ulong tgt_y);
void			gfx_blit(SDL_Surface* source,
					const ulong src_x, const ulong src_y, const ulong src_w, const ulong src_h,
					SDL_Surface* target, const ulong tgt_x, const ulong tgt_y);
void			gfx_clear(SDL_Surface* screen);
void			gfx_flush(void);
void			gfx_set_display(SDL_Surface* screen/*, const fonts font, const font_sizes font_size,
					const int x, const int y, const int grid_w, const int grid_h*/);

SDL_Surface*	gfx_get_resource_buffer(const gfx_resources resource);
int				gfx_get_resource_width(const gfx_resources resource);
int				gfx_get_resource_height(const gfx_resources resource);
void			gfx_load_palette(const gfx_resources resource);

void 			gfx_render_maze(void);
void			gfx_render_keypad(void);
void			gfx_scene_display(void);
void			gfx_inventory_display(void);
void			gfx_menu_display(void);
void			gfx_draw_screen(void);

void			init_framerate(struct FPS_manager* manager);
int				set_framerate(struct FPS_manager* manager, int rate);
int				get_framerate(struct FPS_manager* manager);
void			framerate_delay(struct FPS_manager* manager);

#endif
