#ifndef _util_h_
#define _util_h_

#include <stdlib.h>
#include <stdio.h>

#include <SDL.h>
#include <SDL/SDL_image.h>

#include "config.h"

#ifndef DATADIR
#error "DATADIR must be defined"
#endif

#ifndef SAVESDIR
#error "SAVESDIR must be defined"
#endif

#define AND &&
#define OR ||
#define NOT !
#define NOP {;}
#define true 1
#define false 0
#define MAX(X,Y) ((X)>(Y)?(X):(Y))
#define MIN(X,Y) ((X)<(Y)?(X):(Y))

#define STR_MAX 8192

typedef char* str;
typedef void* ptr;
typedef unsigned char byte;
typedef unsigned int uint;
typedef unsigned long int ulong;
typedef int bool;
typedef FILE* fileHandle;


struct app_options_set
{
	bool	verbose;
	bool	freeze_when_no_focus;
	bool	skip_intro;
	bool	nosound;
};

struct cell
{
	struct cell*	previous;
	ptr				content;
	struct cell*	next;
};

typedef struct cell* p_cell;

struct list
{
	p_cell	head;
	ulong	size;
	p_cell	tail;
};

typedef struct list* p_list;

void	message(const str module, const str type, const str text);
void	info(const str module, const str text);
void	warning(const str module, const str text);
void	error(const str module, const str text);
void	fatal(const str module, const str text);

ptr		xfree(ptr* pointer);
str		strdup(const str text);

void	list_insert(const p_list thelist, const ptr content);
void	list_push(const p_list thelist, const ptr content);
void	list_pop(const p_list thelist);
void	list_clear(const p_list thelist);
void	list_init(const p_list thelist);
void	list_remove(const p_list thelist, const ptr content);
void	list_remove_cell(const p_list thelist, const p_cell thecell);
#ifdef DEBUG
void	list_tests(void);
#endif

bool	file_exists(const str filename);

#endif
